"""checkpoint_1

Revision ID: 9f150e4f9a05
Revises: 53502b6d63bf
Create Date: 2025-05-21 12:25:55.556388

"""
import datetime

from alembic.operations import Operations
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '9f150e4f9a05'
down_revision = '53502b6d63bf'
branch_labels = None
depends_on = None


def upgrade(op: Operations = None):
    # region skip migration if it is existing app, apply if it is new app
    if op is None:
        # 'op' is passed only from migrate.py when applying checkpoin migration
        return
    connection = op.get_bind()
    inspector = sa.inspect(connection)
    if 'alembic_version' in inspector.get_table_names():
        # If version_num exists, then it is existing app
        result = connection.execute(sa.text("SELECT version_num FROM alembic_version"))
        current_version = result.scalar()
        if current_version is not None:
            return
    # endregion

    op.create_table(
        'agents',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('user_class', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('model_name', sa.String(), nullable=True),
        sa.Column('provider', sa.String(), nullable=True),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'chat_bots_history',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('chat_bot_id', sa.Integer(), nullable=False),
        sa.Column('type', sa.String(), nullable=True),
        sa.Column('text', sa.String(), nullable=True),
        sa.Column('user', sa.String(), nullable=True),
        sa.Column('destination', sa.String(), nullable=True),
        sa.Column('sent_at', sa.DateTime(), nullable=True),
        sa.Column('error', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'file',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('source_file_path', sa.String(), nullable=False),
        sa.Column('file_path', sa.String(), nullable=False),
        sa.Column('row_count', sa.Integer(), nullable=False),
        sa.Column('columns', mindsdb.interfaces.storage.db.Json(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('metadata', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_file_name_company_id')
    )
    op.create_table(
        'integration',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('engine', sa.String(), nullable=False),
        sa.Column('data', mindsdb.interfaces.storage.db.Json(), nullable=True),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_integration_name_company_id')
    )
    op.create_table(
        'jobs',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('user_class', sa.Integer(), nullable=True),
        sa.Column('active', sa.Boolean(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('query_str', sa.String(), nullable=False),
        sa.Column('if_query_str', sa.String(), nullable=True),
        sa.Column('start_at', sa.DateTime(), nullable=True),
        sa.Column('end_at', sa.DateTime(), nullable=True),
        sa.Column('next_run_at', sa.DateTime(), nullable=True),
        sa.Column('schedule_str', sa.String(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'jobs_history',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('job_id', sa.Integer(), nullable=True),
        sa.Column('query_str', sa.String(), nullable=True),
        sa.Column('start_at', sa.DateTime(), nullable=True),
        sa.Column('end_at', sa.DateTime(), nullable=True),
        sa.Column('error', sa.String(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('job_id', 'start_at', name='uniq_job_history_job_id_start')
    )
    op.create_table(
        'json_storage',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('resource_group', sa.String(), nullable=True),
        sa.Column('resource_id', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=True),
        sa.Column('content', sa.JSON(), nullable=True),
        sa.Column('encrypted_content', sa.LargeBinary(), nullable=True),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'llm_data',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('input', sa.String(), nullable=False),
        sa.Column('output', sa.String(), nullable=False),
        sa.Column('model_id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'llm_log',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=False),
        sa.Column('api_key', sa.String(), nullable=True),
        sa.Column('model_id', sa.Integer(), nullable=True),
        sa.Column('model_group', sa.String(), nullable=True),
        sa.Column('input', sa.JSON(), nullable=True),
        sa.Column('output', sa.JSON(), nullable=True),
        sa.Column('start_time', sa.DateTime(), nullable=False),
        sa.Column('end_time', sa.DateTime(), nullable=True),
        sa.Column('cost', sa.Numeric(precision=5, scale=2), nullable=True),
        sa.Column('prompt_tokens', sa.Integer(), nullable=True),
        sa.Column('completion_tokens', sa.Integer(), nullable=True),
        sa.Column('total_tokens', sa.Integer(), nullable=True),
        sa.Column('success', sa.Boolean(), nullable=False),
        sa.Column('exception', sa.String(), nullable=True),
        sa.Column('traceback', sa.String(), nullable=True),
        sa.Column('stream', sa.Boolean(), nullable=True, comment="Is this completion done in 'streaming' mode"),
        sa.Column('metadata', sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'project',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('metadata', sa.JSON(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_project_name_company_id')
    )
    op.create_table(
        'queries',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('sql', sa.String(), nullable=False),
        sa.Column('database', sa.String(), nullable=True),
        sa.Column('started_at', sa.DateTime(), nullable=True),
        sa.Column('finished_at', sa.DateTime(), nullable=True),
        sa.Column('parameters', sa.JSON(), nullable=True),
        sa.Column('context', sa.JSON(), nullable=True),
        sa.Column('processed_rows', sa.Integer(), nullable=True),
        sa.Column('error', sa.String(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'query_context',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('query', sa.String(), nullable=False),
        sa.Column('context_name', sa.String(), nullable=False),
        sa.Column('values', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'skills',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('type', sa.String(), nullable=False),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'tasks',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('user_class', sa.Integer(), nullable=True),
        sa.Column('object_type', sa.String(), nullable=False),
        sa.Column('object_id', sa.Integer(), nullable=False),
        sa.Column('last_error', sa.String(), nullable=True),
        sa.Column('active', sa.Boolean(), nullable=True),
        sa.Column('reload', sa.Boolean(), nullable=True),
        sa.Column('run_by', sa.String(), nullable=True),
        sa.Column('alive_time', sa.DateTime(timezone=True), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'triggers',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('database_id', sa.Integer(), nullable=False),
        sa.Column('table_name', sa.String(), nullable=False),
        sa.Column('query_str', sa.String(), nullable=False),
        sa.Column('columns', sa.String(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'agent_skills',
        sa.Column('agent_id', sa.Integer(), nullable=False),
        sa.Column('skill_id', sa.Integer(), nullable=False),
        sa.Column('parameters', sa.JSON(), nullable=True),
        sa.ForeignKeyConstraint(['agent_id'], ['agents.id'], ),
        sa.ForeignKeyConstraint(['skill_id'], ['skills.id'], ),
        sa.PrimaryKeyConstraint('agent_id', 'skill_id')
    )
    op.create_table(
        'chat_bots',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('agent_id', sa.Integer(), nullable=True),
        sa.Column('model_name', sa.String(), nullable=True),
        sa.Column('database_id', sa.Integer(), nullable=True),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('webhook_token', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(['agent_id'], ['agents.id'], name='fk_agent_id'),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'predictor',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('deleted_at', sa.DateTime(), nullable=True),
        sa.Column('name', sa.String(), nullable=True),
        sa.Column('data', mindsdb.interfaces.storage.db.Json(), nullable=True),
        sa.Column('to_predict', mindsdb.interfaces.storage.db.Array(), nullable=True),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('mindsdb_version', sa.String(), nullable=True),
        sa.Column('native_version', sa.String(), nullable=True),
        sa.Column('integration_id', sa.Integer(), nullable=True),
        sa.Column('data_integration_ref', mindsdb.interfaces.storage.db.Json(), nullable=True),
        sa.Column('fetch_data_query', sa.String(), nullable=True),
        sa.Column('learn_args', mindsdb.interfaces.storage.db.Json(), nullable=True),
        sa.Column('update_status', sa.String(), nullable=True),
        sa.Column('status', sa.String(), nullable=True),
        sa.Column('active', sa.Boolean(), nullable=True),
        sa.Column('training_data_columns_count', sa.Integer(), nullable=True),
        sa.Column('training_data_rows_count', sa.Integer(), nullable=True),
        sa.Column('training_start_at', sa.DateTime(), nullable=True),
        sa.Column('training_stop_at', sa.DateTime(), nullable=True),
        sa.Column('label', sa.String(), nullable=True),
        sa.Column('version', sa.Integer(), nullable=True),
        sa.Column('code', sa.String(), nullable=True),
        sa.Column('lightwood_version', sa.String(), nullable=True),
        sa.Column('dtype_dict', mindsdb.interfaces.storage.db.Json(), nullable=True),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('training_phase_current', sa.Integer(), nullable=True),
        sa.Column('training_phase_total', sa.Integer(), nullable=True),
        sa.Column('training_phase_name', sa.String(), nullable=True),
        sa.Column('training_metadata', sa.JSON(), nullable=False),
        sa.ForeignKeyConstraint(['integration_id'], ['integration.id'], name='fk_integration_id'),
        sa.ForeignKeyConstraint(['project_id'], ['project.id'], name='fk_project_id'),
        sa.PrimaryKeyConstraint('id')
    )
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.create_index('predictor_index', ['company_id', 'name', 'version', 'active', 'deleted_at'], unique=True)

    op.create_table(
        'view',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('query', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['project_id'], ['project.id'], name='fk_project_id'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'company_id', name='unique_view_name_company_id')
    )
    op.create_table(
        'knowledge_base',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('project_id', sa.Integer(), nullable=False),
        sa.Column('params', sa.JSON(), nullable=True),
        sa.Column('vector_database_id', sa.Integer(), nullable=True),
        sa.Column('vector_database_table', sa.String(), nullable=True),
        sa.Column('embedding_model_id', sa.Integer(), nullable=True),
        sa.Column('query_id', sa.Integer(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['embedding_model_id'], ['predictor.id'], name='fk_knowledge_base_embedding_model_id'),
        sa.ForeignKeyConstraint(['vector_database_id'], ['integration.id'], name='fk_knowledge_base_vector_database_id'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'project_id', name='unique_knowledge_base_name_project_id')
    )

    # Insert default project
    op.bulk_insert(
        sa.table(
            'project',
            sa.Column('name', sa.String()),
            sa.Column('company_id', sa.Integer()),
            sa.Column('metadata', sa.JSON()),
            sa.Column('created_at', sa.DateTime())
        ),
        [{
            'name': 'mindsdb',
            'company_id': 0,
            'metadata': {'is_default': True},
            'created_at': datetime.datetime.now()
        }]
    )


def downgrade():
    # do nothging, since it is checkpoint migration
    pass



import inspect
from collections import defaultdict

from flask import current_app as ca, request
from flask_restx import Resource

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.tree import ns_conf
from mindsdb.metrics.metrics import api_endpoint_metrics


@ns_conf.route("/")
class GetRoot(Resource):
    @ns_conf.doc("get_tree_root")
    @api_endpoint_metrics("GET", "/tree")
    def get(self):
        databases = ca.database_controller.get_list()
        result = [
            {
                "name": x["name"],
                "class": "db",
                "type": x["type"],
                "engine": x["engine"],
                "deletable": x["deletable"],
                "visible": x["visible"],
            }
            for x in databases
        ]
        return result


@ns_conf.route("/<db_name>")
@ns_conf.param("db_name", "Name of the database")
class GetLeaf(Resource):
    @ns_conf.doc("get_tree_leaf")
    @api_endpoint_metrics("GET", "/tree/database")
    def get(self, db_name):
        with_schemas = request.args.get("all_schemas")
        if isinstance(with_schemas, str):
            with_schemas = with_schemas.lower() in ("1", "true")
        else:
            with_schemas = False
        db_name = db_name.lower()
        databases = ca.database_controller.get_dict()
        if db_name not in databases:
            return http_error(400, "Error", f"There is no element with name '{db_name}'")
        db = databases[db_name]
        if db["type"] == "project":
            project = ca.database_controller.get_project(db_name)
            tables = project.get_tables()
            tables = [
                {
                    "name": key,
                    "schema": None,
                    "class": "table",
                    "type": val["type"],
                    "engine": val.get("engine"),
                    "deletable": val.get("deletable"),
                }
                for key, val in tables.items()
            ]

            jobs = ca.jobs_controller.get_list(db_name)
            tables = tables + [
                {"name": job["name"], "schema": None, "class": "job", "type": "job", "engine": "job", "deletable": True}
                for job in jobs
            ]
        elif db["type"] == "data":
            handler = ca.integration_controller.get_data_handler(db_name)
            if "all" in inspect.signature(handler.get_tables).parameters:
                response = handler.get_tables(all=with_schemas)
            else:
                response = handler.get_tables()
            if response.type != "table":
                return []
            table_types = {"BASE TABLE": "table", "VIEW": "view"}
            tables = response.data_frame.to_dict(orient="records")

            schemas = defaultdict(list)

            for table_meta in tables:
                table_meta = {key.lower(): val for key, val in table_meta.items()}
                schama = table_meta.get("table_schema")
                schemas[schama].append(
                    {
                        "name": table_meta["table_name"],
                        "class": "table",
                        "type": table_types.get(table_meta.get("table_type")),
                        "engine": None,
                        "deletable": False,
                    }
                )
            if len(schemas) == 1 and list(schemas.keys())[0] is None:
                tables = schemas[None]
            else:
                tables = [
                    {"name": key, "class": "schema", "deletable": False, "children": val}
                    for key, val in schemas.items()
                ]
        elif db["type"] == "system":
            system_db = ca.database_controller.get_system_db(db_name)
            tables = system_db.get_tree_tables()
            tables = [
                {
                    "name": table.name,
                    "class": table.kind,
                    "type": "system view",
                    "engine": None,
                    "deletable": table.deletable,
                }
                for table in tables.values()
            ]
        return tables



import time
from typing import List

import pandas as pd

from mindsdb.integrations.handlers.rag_handler.settings import (
    PersistedVectorStoreSaver,
    PersistedVectorStoreSaverConfig,
    RAGBaseParameters,
    VectorStoreFactory,
    df_to_documents,
    get_chroma_client,
    load_embeddings_model,
    url_to_documents,
)
from mindsdb.utilities import log
from langchain_core.documents import Document
from langchain_core.vectorstores import VectorStore
from langchain_text_splitters import RecursiveCharacterTextSplitter

logger = log.getLogger(__name__)


def validate_document(doc) -> bool:
    """Check an individual document."""
    # Example checks
    if not isinstance(doc, Document):
        return False

    if not doc.page_content:
        return False

    return True


def validate_documents(documents) -> bool:
    """Validate document list format."""

    if not isinstance(documents, list):
        return False

    if not documents:
        return False

    # Check fields/format of a document
    return all([validate_document(doc) for doc in documents])


class RAGIngestor:
    """A class for converting a dataframe and/or url to a vectorstore embedded with a given embeddings model"""

    def __init__(
        self,
        args: RAGBaseParameters,
        df: pd.DataFrame,
    ):
        self.args = args
        self.df = df
        self.embeddings_model_name = args.embeddings_model_name

        self.vector_store = VectorStoreFactory.get_vectorstore_class(
            args.vector_store_name
        )

    def split_documents(self, chunk_size, chunk_overlap) -> list:
        # Load documents and split in chunks
        logger.info("Loading documents from input data")

        documents = []

        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size, chunk_overlap=chunk_overlap
        )

        if self.df is not None:
            # if user provides a dataframe, load documents from dataframe
            documents.extend(
                df_to_documents(
                    df=self.df,
                    page_content_columns=self.args.context_columns,
                    url_column_name=self.args.url_column_name,
                )
            )

        if self.args.url:
            # if user provides a url, load documents from url
            documents.extend(url_to_documents(self.args.url))

        n_tokens = sum([len(doc.page_content) for doc in documents])

        # split documents into chunks of text
        texts = text_splitter.split_documents(documents)
        logger.info(f"Loaded {len(documents)} documents from input data")
        logger.info(f"Total number of tokens: {n_tokens}")
        logger.info(f"Split into {len(texts)} chunks of text (tokens)")

        return texts

    def create_db_from_documents(self, documents, embeddings_model) -> VectorStore:
        """Create DB from documents."""

        if self.args.vector_store_name == "chromadb":

            return self.vector_store.from_documents(
                documents=documents,
                embedding=embeddings_model,
                client=get_chroma_client(
                    persist_directory=self.args.vector_store_storage_path
                ),
                collection_name=self.args.collection_name,
            )
        else:
            return self.create_db_from_texts(documents, embeddings_model)

    def create_db_from_texts(self, documents, embeddings_model) -> VectorStore:
        """Create DB from text content."""

        texts = [doc.page_content for doc in documents]
        metadata = [doc.metadata for doc in documents]

        return self.vector_store.from_texts(
            texts=texts, embedding=embeddings_model, metadatas=metadata
        )

    @staticmethod
    def _create_batch_embeddings(documents: List[Document], embeddings_batch_size):
        """
        create batch of document embeddings
        """

        for i in range(0, len(documents), embeddings_batch_size):
            yield documents[i: i + embeddings_batch_size]

    def embeddings_to_vectordb(self) -> None:
        """Create vectorstore from documents and store locally."""

        start_time = time.time()

        # Load documents and splits in chunks (if not in evaluation_type mode)
        documents = self.split_documents(
            chunk_size=self.args.chunk_size, chunk_overlap=self.args.chunk_overlap
        )

        # Load embeddings model
        embeddings_model = load_embeddings_model(
            self.embeddings_model_name, self.args.use_gpu
        )

        logger.info("Creating vectorstore from documents")

        if not validate_documents(documents):
            raise ValueError("Invalid documents")

        try:
            db = self.create_db_from_documents(documents, embeddings_model)
        except Exception as e:
            raise Exception(
                f"Error loading embeddings to {self.args.vector_store_name}: {e}"
            )

        config = PersistedVectorStoreSaverConfig(
            vector_store_name=self.args.vector_store_name,
            vector_store=db,
            persist_directory=self.args.vector_store_storage_path,
            collection_name=self.args.collection_name,
        )

        vector_store_saver = PersistedVectorStoreSaver(config)

        vector_store_saver.save_vector_store(db)

        db = None  # Free up memory

        end_time = time.time()
        elapsed_time = round(end_time - start_time)

        logger.info(f"Finished creating {self.args.vector_store_name} from texts, it has been "
                    f"persisted to {self.args.vector_store_storage_path}")

        time_minutes = round(elapsed_time / 60)

        if time_minutes > 1:
            logger.info(f"Elapsed time: {time_minutes} minutes")
        else:
            logger.info(f"Elapsed time: {elapsed_time} seconds")



import copy
import shutil
import tempfile
from pathlib import Path
from http import HTTPStatus

from flask import request
from flask_restx import Resource
from flask import current_app as ca

from mindsdb.api.http.namespaces.configs.config import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities import log
from mindsdb.utilities.functions import decrypt, encrypt
from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.response import HandlerStatusResponse


logger = log.getLogger(__name__)


@ns_conf.route("/")
@ns_conf.param("name", "Get config")
class GetConfig(Resource):
    @ns_conf.doc("get_config")
    @api_endpoint_metrics("GET", "/config")
    def get(self):
        config = Config()
        resp = {"auth": {"http_auth_enabled": config["auth"]["http_auth_enabled"]}}
        for key in ["default_llm", "default_embedding_model", "default_reranking_model"]:
            value = config.get(key)
            if value is not None:
                resp[key] = value
        return resp

    @ns_conf.doc("put_config")
    @api_endpoint_metrics("PUT", "/config")
    def put(self):
        data = request.json

        allowed_arguments = {"auth", "default_llm", "default_embedding_model", "default_reranking_model"}
        unknown_arguments = list(set(data.keys()) - allowed_arguments)
        if len(unknown_arguments) > 0:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong arguments", f"Unknown argumens: {unknown_arguments}")

        nested_keys_to_validate = {"auth"}
        for key in data.keys():
            if key in nested_keys_to_validate:
                unknown_arguments = list(set(data[key].keys()) - set(Config()[key].keys()))
                if len(unknown_arguments) > 0:
                    return http_error(
                        HTTPStatus.BAD_REQUEST, "Wrong arguments", f"Unknown argumens: {unknown_arguments}"
                    )

        overwrite_arguments = {"default_llm", "default_embedding_model", "default_reranking_model"}
        overwrite_data = {k: data[k] for k in overwrite_arguments if k in data}
        merge_data = {k: data[k] for k in data if k not in overwrite_arguments}

        if len(overwrite_data) > 0:
            Config().update(overwrite_data, overwrite=True)
        if len(merge_data) > 0:
            Config().update(merge_data)

        Config().update(data)

        return "", 200


@ns_conf.route("/integrations")
@ns_conf.param("name", "List all database integration")
class ListIntegration(Resource):
    @api_endpoint_metrics("GET", "/config/integrations")
    def get(self):
        return {"integrations": [k for k in ca.integration_controller.get_all(show_secrets=False)]}


@ns_conf.route("/all_integrations")
@ns_conf.param("name", "List all database integration")
class AllIntegration(Resource):
    @ns_conf.doc("get_all_integrations")
    @api_endpoint_metrics("GET", "/config/all_integrations")
    def get(self):
        integrations = ca.integration_controller.get_all(show_secrets=False)
        return integrations


@ns_conf.route("/integrations/<name>")
@ns_conf.param("name", "Database integration")
class Integration(Resource):
    @ns_conf.doc("get_integration")
    @api_endpoint_metrics("GET", "/config/integrations/integration")
    def get(self, name):
        integration = ca.integration_controller.get(name, show_secrets=False)
        if integration is None:
            return http_error(HTTPStatus.NOT_FOUND, "Not found", f"Can't find integration: {name}")
        integration = copy.deepcopy(integration)
        return integration

    @ns_conf.doc("put_integration")
    @api_endpoint_metrics("PUT", "/config/integrations/integration")
    def put(self, name):
        params = {}
        if request.is_json:
            params.update((request.json or {}).get("params", {}))
        else:
            params.update(request.form or {})

        if len(params) == 0:
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", "type of 'params' must be dict")

        files = request.files
        temp_dir = None
        if files is not None and len(files) > 0:
            temp_dir = tempfile.mkdtemp(prefix="integration_files_")
            for key, file in files.items():
                temp_dir_path = Path(temp_dir)
                file_name = Path(file.filename)
                file_path = temp_dir_path.joinpath(file_name).resolve()
                if temp_dir_path not in file_path.parents:
                    raise Exception(f"Can not save file at path: {file_path}")
                file.save(file_path)
                params[key] = str(file_path)

        is_test = params.get("test", False)
        # TODO: Move this to new Endpoint

        config = Config()
        secret_key = config.get("secret_key", "dummy-key")

        if is_test:
            del params["test"]
            handler_type = params.pop("type", None)
            params.pop("publish", None)
            try:
                handler = ca.integration_controller.create_tmp_handler(name, handler_type, params)
                status = handler.check_connection()
            except ImportError as e:
                status = HandlerStatusResponse(success=False, error_message=str(e))
            if temp_dir is not None:
                shutil.rmtree(temp_dir)

            resp = status.to_json()

            if status.success and "code" in params:
                if hasattr(handler, "handler_storage"):
                    # attach storage if exists
                    export = handler.handler_storage.export_files()
                    if export:
                        # encrypt with flask secret key
                        encrypted = encrypt(export, secret_key)
                        resp["storage"] = encrypted.decode()

            return resp, 200

        config = Config()
        secret_key = config.get("secret_key", "dummy-key")

        integration = ca.integration_controller.get(name, show_secrets=False)
        if integration is not None:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Wrong argument", f"Integration with name '{name}' already exists"
            )

        try:
            engine = params["type"]
            if engine is not None:
                del params["type"]
            params.pop("publish", False)
            storage = params.pop("storage", None)
            ca.integration_controller.add(name, engine, params)

            # copy storage
            if storage is not None:
                handler = ca.integration_controller.get_data_handler(name)

                export = decrypt(storage.encode(), secret_key)
                handler.handler_storage.import_files(export)

        except Exception as e:
            logger.error(str(e))
            if temp_dir is not None:
                shutil.rmtree(temp_dir)
            return http_error(HTTPStatus.INTERNAL_SERVER_ERROR, "Error", f"Error during config update: {str(e)}")

        if temp_dir is not None:
            shutil.rmtree(temp_dir)
        return {}, 200

    @ns_conf.doc("delete_integration")
    @api_endpoint_metrics("DELETE", "/config/integrations/integration")
    def delete(self, name):
        integration = ca.integration_controller.get(name)
        if integration is None:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Integration does not exists", f"Nothing to delete. '{name}' not exists."
            )
        try:
            ca.integration_controller.delete(name)
        except Exception as e:
            logger.error(str(e))
            return http_error(HTTPStatus.INTERNAL_SERVER_ERROR, "Error", f"Error during integration delete: {str(e)}")
        return "", 200

    @ns_conf.doc("modify_integration")
    @api_endpoint_metrics("POST", "/config/integrations/integration")
    def post(self, name):
        params = {}
        params.update((request.json or {}).get("params", {}))
        params.update(request.form or {})

        if not isinstance(params, dict):
            return http_error(HTTPStatus.BAD_REQUEST, "Wrong argument", "type of 'params' must be dict")
        integration = ca.integration_controller.get(name)
        if integration is None:
            return http_error(
                HTTPStatus.BAD_REQUEST, "Integration does not exists", f"Nothin to modify. '{name}' not exists."
            )
        try:
            if "enabled" in params:
                params["publish"] = params["enabled"]
                del params["enabled"]
            ca.integration_controller.modify(name, params)

        except Exception as e:
            logger.error(str(e))
            return http_error(
                HTTPStatus.INTERNAL_SERVER_ERROR, "Error", f"Error during integration modification: {str(e)}"
            )
        return "", 200



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        "type": ARG_TYPE.STR,
        "description": "The hostname, IP address, or URL of the IBM Db2 database.",
        "required": True,
        "label": "Host"
    },
    database={
        "type": ARG_TYPE.STR,
        "description": "The name of the IBM Db2 database to connect to.",
        "required": True,
        "label": "Database"
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The username for the IBM Db2 database.",
        "required": True,
        "label": "User"
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password for the IBM Db2 database.",
        'secret': True,
        "required": True,
        "label": "Password"
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "The port number for connecting to the IBM Db2 database. Default is `50000`",
        "required": False,
        "label": "Port"
    },
    schema={
        "type": ARG_TYPE.STR,
        "description": "The database schema to use within the IBM Db2 database.",
        "required": False,
        "label": "Schema"
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1",
    port="25000",
    password="1234",
    user="db2admin",
    schema="db2admin",
    database="BOOKS",
)



import ast
import hashlib
from enum import Enum
from typing import Dict, List, Optional
import datetime as dt

import pandas as pd
from mindsdb_sql_parser.ast import (
    BinaryOperation,
    Constant,
    CreateTable,
    Delete,
    DropTables,
    Insert,
    Select,
    Star,
    Tuple,
    Update,
)
from mindsdb_sql_parser.ast.base import ASTNode

from mindsdb.integrations.libs.response import RESPONSE_TYPE, HandlerResponse
from mindsdb.utilities import log
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator, KeywordSearchArgs

from mindsdb.integrations.utilities.query_traversal import query_traversal
from .base import BaseHandler

LOG = log.getLogger(__name__)


class VectorHandlerException(Exception): ...


class TableField(Enum):
    """
    Enum for table fields.
    """

    ID = "id"
    CONTENT = "content"
    EMBEDDINGS = "embeddings"
    METADATA = "metadata"
    SEARCH_VECTOR = "search_vector"
    DISTANCE = "distance"
    RELEVANCE = "relevance"


class DistanceFunction(Enum):
    SQUARED_EUCLIDEAN_DISTANCE = ("<->",)
    NEGATIVE_DOT_PRODUCT = ("<#>",)
    COSINE_DISTANCE = "<=>"


class VectorStoreHandler(BaseHandler):
    """
    Base class for handlers associated to vector databases.
    """

    SCHEMA = [
        {
            "name": TableField.ID.value,
            "data_type": "string",
        },
        {
            "name": TableField.CONTENT.value,
            "data_type": "string",
        },
        {
            "name": TableField.EMBEDDINGS.value,
            "data_type": "list",
        },
        {
            "name": TableField.METADATA.value,
            "data_type": "json",
        },
        {
            "name": TableField.DISTANCE.value,
            "data_type": "float",
        },
    ]

    def validate_connection_parameters(self, name, **kwargs):
        """Create validation for input parameters."""

        return NotImplementedError()

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def disconnect(self):
        pass

    def _value_or_self(self, value):
        if isinstance(value, Constant):
            return value.value
        else:
            return value

    def extract_conditions(self, where_statement) -> Optional[List[FilterCondition]]:
        conditions = []
        # parse conditions
        if where_statement is not None:
            # dfs to get all binary operators in the where statement
            def _extract_comparison_conditions(node, **kwargs):
                if isinstance(node, BinaryOperation):
                    # if the op is and, continue
                    # TODO: need to handle the OR case
                    if node.op.upper() == "AND":
                        return
                    op = FilterOperator(node.op.upper())
                    # unquote the left hand side
                    left_hand = node.args[0].parts[-1].strip("`")
                    if isinstance(node.args[1], Constant):
                        if left_hand == TableField.SEARCH_VECTOR.value:
                            right_hand = ast.literal_eval(node.args[1].value)
                        else:
                            right_hand = node.args[1].value
                    elif isinstance(node.args[1], Tuple):
                        # Constant could be actually a list i.e. [1.2, 3.2]
                        right_hand = [item.value for item in node.args[1].items]
                    else:
                        raise Exception(f"Unsupported right hand side: {node.args[1]}")
                    conditions.append(FilterCondition(column=left_hand, op=op, value=right_hand))

            query_traversal(where_statement, _extract_comparison_conditions)

        else:
            conditions = None

        return conditions

    def _convert_metadata_filters(self, conditions, allowed_metadata_columns=None):
        if conditions is None:
            return
        # try to treat conditions that are not in TableField as metadata conditions
        for condition in conditions:
            if self._is_metadata_condition(condition):
                # check restriction
                if allowed_metadata_columns is not None:
                    # system columns are underscored, skip them
                    if condition.column.lower() not in allowed_metadata_columns and not condition.column.startswith(
                        "_"
                    ):
                        raise ValueError(f"Column is not found: {condition.column}")

                # convert if required
                if not condition.column.startswith(TableField.METADATA.value):
                    condition.column = TableField.METADATA.value + "." + condition.column

    def _is_columns_allowed(self, columns: List[str]) -> bool:
        """
        Check if columns are allowed.
        """
        allowed_columns = set([col["name"] for col in self.SCHEMA])
        return set(columns).issubset(allowed_columns)

    def _is_metadata_condition(self, condition: FilterCondition) -> bool:
        allowed_field_values = set([field.value for field in TableField])
        if condition.column in allowed_field_values:
            return False
        return True

    def _dispatch_create_table(self, query: CreateTable):
        """
        Dispatch create table query to the appropriate method.
        """
        # parse key arguments
        table_name = query.name.parts[-1]
        if_not_exists = getattr(query, "if_not_exists", False)
        return self.create_table(table_name, if_not_exists=if_not_exists)

    def _dispatch_drop_table(self, query: DropTables):
        """
        Dispatch drop table query to the appropriate method.
        """
        table_name = query.tables[0].parts[-1]
        if_exists = getattr(query, "if_exists", False)

        return self.drop_table(table_name, if_exists=if_exists)

    def _dispatch_insert(self, query: Insert):
        """
        Dispatch insert query to the appropriate method.
        """
        # parse key arguments
        table_name = query.table.parts[-1]
        columns = [column.name for column in query.columns]

        if not self._is_columns_allowed(columns):
            raise Exception(f"Columns {columns} not allowed.Allowed columns are {[col['name'] for col in self.SCHEMA]}")

        # get content column if it is present
        if TableField.CONTENT.value in columns:
            content_col_index = columns.index("content")
            content = [self._value_or_self(row[content_col_index]) for row in query.values]
        else:
            content = None

        # get id column if it is present
        ids = None
        if TableField.ID.value in columns:
            id_col_index = columns.index("id")
            ids = [self._value_or_self(row[id_col_index]) for row in query.values]
        elif TableField.CONTENT.value is None:
            raise Exception("Content or id is required!")

        # get embeddings column if it is present
        if TableField.EMBEDDINGS.value in columns:
            embeddings_col_index = columns.index("embeddings")
            embeddings = [ast.literal_eval(self._value_or_self(row[embeddings_col_index])) for row in query.values]
        else:
            raise Exception("Embeddings column is required!")

        if TableField.METADATA.value in columns:
            metadata_col_index = columns.index("metadata")
            metadata = [ast.literal_eval(self._value_or_self(row[metadata_col_index])) for row in query.values]
        else:
            metadata = None

        # create dataframe
        data = {
            TableField.CONTENT.value: content,
            TableField.EMBEDDINGS.value: embeddings,
            TableField.METADATA.value: metadata,
        }
        if ids is not None:
            data[TableField.ID.value] = ids

        return self.do_upsert(table_name, pd.DataFrame(data))

    def dispatch_update(self, query: Update, conditions: List[FilterCondition] = None):
        """
        Dispatch update query to the appropriate method.
        """
        table_name = query.table.parts[-1]

        row = {}
        for k, v in query.update_columns.items():
            k = k.lower()
            if isinstance(v, Constant):
                v = v.value
            if k == TableField.EMBEDDINGS.value and isinstance(v, str):
                # it could be embeddings in string
                try:
                    v = ast.literal_eval(v)
                except Exception:
                    pass
            row[k] = v

        if conditions is None:
            where_statement = query.where
            conditions = self.extract_conditions(where_statement)

        for condition in conditions:
            if condition.op != FilterOperator.EQUAL:
                raise NotImplementedError

            row[condition.column] = condition.value

        # checks
        if TableField.EMBEDDINGS.value not in row:
            raise Exception("Embeddings column is required!")

        if TableField.CONTENT.value not in row:
            raise Exception("Content is required!")

        # store
        df = pd.DataFrame([row])

        return self.do_upsert(table_name, df)

    def set_metadata_cur_time(self, df, col_name):
        metadata_col = TableField.METADATA.value
        cur_date = dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        def set_time(meta):
            meta[col_name] = cur_date

        df[metadata_col].apply(set_time)

    def do_upsert(self, table_name, df):
        """Upsert data into table, handling document updates and deletions.

        Args:
            table_name (str): Name of the table
            df (pd.DataFrame): DataFrame containing the data to upsert

        The function handles three cases:
        1. New documents: Insert them
        2. Updated documents: Delete old chunks and insert new ones
        """
        id_col = TableField.ID.value
        metadata_col = TableField.METADATA.value
        content_col = TableField.CONTENT.value

        def gen_hash(v):
            return hashlib.md5(str(v).encode()).hexdigest()

        if id_col not in df.columns:
            # generate for all
            df[id_col] = df[content_col].apply(gen_hash)
        else:
            # generate for empty
            for i in range(len(df)):
                if pd.isna(df.loc[i, id_col]):
                    df.loc[i, id_col] = gen_hash(df.loc[i, content_col])

        # remove duplicated ids
        df = df.drop_duplicates([TableField.ID.value])

        # id is string TODO is it ok?
        df[id_col] = df[id_col].apply(str)

        # set updated_at
        self.set_metadata_cur_time(df, "_updated_at")

        if hasattr(self, "upsert"):
            self.upsert(table_name, df)
            return

        # find existing ids
        df_existed = self.select(
            table_name,
            columns=[id_col, metadata_col],
            conditions=[FilterCondition(column=id_col, op=FilterOperator.IN, value=list(df[id_col]))],
        )
        existed_ids = list(df_existed[id_col])

        # update existed
        df_update = df[df[id_col].isin(existed_ids)]
        df_insert = df[~df[id_col].isin(existed_ids)]

        if not df_update.empty:
            # get values of existed `created_at` and return them to metadata
            origin_id_col = "_original_doc_id"

            created_dates, ids = {}, {}
            for _, row in df_existed.iterrows():
                chunk_id = row[id_col]
                created_dates[chunk_id] = row[metadata_col].get("_created_at")
                ids[chunk_id] = row[metadata_col].get(origin_id_col)

            def keep_created_at(row):
                val = created_dates.get(row[id_col])
                if val:
                    row[metadata_col]["_created_at"] = val
                # keep id column
                if origin_id_col not in row[metadata_col]:
                    row[metadata_col][origin_id_col] = ids.get(row[id_col])
                return row

            df_update.apply(keep_created_at, axis=1)

            try:
                self.update(table_name, df_update, [id_col])
            except NotImplementedError:
                # not implemented? do it with delete and insert
                conditions = [FilterCondition(column=id_col, op=FilterOperator.IN, value=list(df[id_col]))]
                self.delete(table_name, conditions)
                self.insert(table_name, df_update)
        if not df_insert.empty:
            # set created_at
            self.set_metadata_cur_time(df_insert, "_created_at")

            self.insert(table_name, df_insert)

    def dispatch_delete(self, query: Delete, conditions: List[FilterCondition] = None):
        """
        Dispatch delete query to the appropriate method.
        """
        # parse key arguments
        table_name = query.table.parts[-1]
        if conditions is None:
            where_statement = query.where
            conditions = self.extract_conditions(where_statement)
        self._convert_metadata_filters(conditions)

        # dispatch delete
        return self.delete(table_name, conditions=conditions)

    def dispatch_select(
        self,
        query: Select,
        conditions: Optional[List[FilterCondition]] = None,
        allowed_metadata_columns: List[str] = None,
        keyword_search_args: Optional[KeywordSearchArgs] = None,
    ):
        """
        Dispatches a select query to the appropriate method, handling both
        standard selections and keyword searches based on the provided arguments.
        """
        # 1. Parse common query arguments
        table_name = query.from_table.parts[-1]

        # If targets are a star (*), select all schema columns
        if isinstance(query.targets[0], Star):
            columns = [col["name"] for col in self.SCHEMA]
        else:
            columns = [col.parts[-1] for col in query.targets]

        # 2. Validate columns
        if not self._is_columns_allowed(columns):
            allowed_cols = [col["name"] for col in self.SCHEMA]
            raise Exception(f"Columns {columns} not allowed. Allowed columns are {allowed_cols}")

        # 3. Extract and process conditions
        if conditions is None:
            where_statement = query.where
            conditions = self.extract_conditions(where_statement)
        self._convert_metadata_filters(conditions, allowed_metadata_columns=allowed_metadata_columns)

        # 4. Get offset and limit
        offset = query.offset.value if query.offset is not None else None
        limit = query.limit.value if query.limit is not None else None

        # 5. Conditionally dispatch to the correct select method
        if keyword_search_args:
            # It's a keyword search
            return self.keyword_select(
                table_name,
                columns=columns,
                conditions=conditions,
                offset=offset,
                limit=limit,
                keyword_search_args=keyword_search_args,
            )
        else:
            # It's a standard select
            try:
                return self.select(
                    table_name,
                    columns=columns,
                    conditions=conditions,
                    offset=offset,
                    limit=limit,
                )

            except Exception as e:
                handler_engine = self.__class__.name
                raise VectorHandlerException(f"Error in {handler_engine} database: {e}")

    def _dispatch(self, query: ASTNode) -> HandlerResponse:
        """
        Parse and Dispatch query to the appropriate method.
        """
        dispatch_router = {
            CreateTable: self._dispatch_create_table,
            DropTables: self._dispatch_drop_table,
            Insert: self._dispatch_insert,
            Update: self.dispatch_update,
            Delete: self.dispatch_delete,
            Select: self.dispatch_select,
        }
        if type(query) in dispatch_router:
            resp = dispatch_router[type(query)](query)
            if resp is not None:
                return HandlerResponse(resp_type=RESPONSE_TYPE.TABLE, data_frame=resp)
            else:
                return HandlerResponse(resp_type=RESPONSE_TYPE.OK)

        else:
            raise NotImplementedError(f"Query type {type(query)} not implemented.")

    def query(self, query: ASTNode) -> HandlerResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc

        Returns:
            HandlerResponse
        """
        return self._dispatch(query)

    def create_table(self, table_name: str, if_not_exists=True) -> HandlerResponse:
        """Create table

        Args:
            table_name (str): table name
            if_not_exists (bool): if True, do nothing if table exists

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def drop_table(self, table_name: str, if_exists=True) -> HandlerResponse:
        """Drop table

        Args:
            table_name (str): table name
            if_exists (bool): if True, do nothing if table does not exist

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def insert(self, table_name: str, data: pd.DataFrame) -> HandlerResponse:
        """Insert data into table

        Args:
            table_name (str): table name
            data (pd.DataFrame): data to insert
            columns (List[str]): columns to insert

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def update(self, table_name: str, data: pd.DataFrame, key_columns: List[str] = None):
        """Update data in table

        Args:
            table_name (str): table name
            data (pd.DataFrame): data to update
            key_columns (List[str]): key to  to update

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def delete(self, table_name: str, conditions: List[FilterCondition] = None) -> HandlerResponse:
        """Delete data from table

        Args:
            table_name (str): table name
            conditions (List[FilterCondition]): conditions to delete

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def select(
        self,
        table_name: str,
        columns: List[str] = None,
        conditions: List[FilterCondition] = None,
        offset: int = None,
        limit: int = None,
    ) -> pd.DataFrame:
        """Select data from table

        Args:
            table_name (str): table name
            columns (List[str]): columns to select
            conditions (List[FilterCondition]): conditions to select

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def get_columns(self, table_name: str) -> HandlerResponse:
        # return a fixed set of columns
        data = pd.DataFrame(self.SCHEMA)
        data.columns = ["COLUMN_NAME", "DATA_TYPE"]
        return HandlerResponse(
            resp_type=RESPONSE_TYPE.DATA,
            data_frame=data,
        )

    def hybrid_search(
        self,
        table_name: str,
        embeddings: List[float],
        query: str = None,
        metadata: Dict[str, str] = None,
        distance_function=DistanceFunction.COSINE_DISTANCE,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Executes a hybrid search, combining semantic search and one or both of keyword/metadata search.

        For insight on the query construction, see: https://docs.pgvecto.rs/use-case/hybrid-search.html#advanced-search-merge-the-results-of-full-text-search-and-vector-search.

        Args:
            table_name(str): Name of underlying table containing content, embeddings, & metadata
            embeddings(List[float]): Embedding vector to perform semantic search against
            query(str): User query to convert into keywords for keyword search
            metadata(Dict[str, str]): Metadata filters to filter content rows against
            distance_function(DistanceFunction): Distance function used to compare embeddings vectors for semantic search

        Returns:
            df(pd.DataFrame): Hybrid search result, sorted by hybrid search rank
        """
        raise NotImplementedError(f"Hybrid search not supported for VectorStoreHandler {self.name}")

    def create_index(self, *args, **kwargs):
        """
        Create an index on the specified table.
        """
        raise NotImplementedError(f"create_index not supported for VectorStoreHandler {self.name}")



from collections import OrderedDict

from mindsdb.integrations.handlers.postgres_handler.postgres_handler import (
    PostgresHandler,
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class YugabyteHandler(PostgresHandler):
    """
    This handler handles connection and execution of the YugabyteSQL statements.
    """

    name = 'yugabyte'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the YugabyteDB server/database.',
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the YugabyteDB server.',
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the YugabyteDB server.',
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect YugabyteDB server',
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name  to connect YugabyteDB server',
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': '(OPTIONAL) comma seperated value of schema to be considered while querying',
    },
    sslmode={
        'type': ARG_TYPE.STR,
        'description': ''' (OPTIONAL) Specifies the SSL mode for the connection, determining whether to use SSL encryption and the level of verification required
        (e.g., "**disable**", "**allow**", "**prefer**", "**require**", "**verify-ca**", "**verify-full**")''',
    },
)

connection_args_example = OrderedDict(
    host='127.0.0.1', port=5433, password='', user='admin', database='yugabyte'
)



from mindsdb.integrations.handlers.scylla_handler import Handler as ScyllaHandler


class DatastaxHandler(ScyllaHandler):
    """
    This handler handles connection and execution of the Datastax Astra DB statements.
    """

    name = "astra"

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



from io import BytesIO
import os
from typing import Union
from urllib.parse import urlparse
import xml.etree.ElementTree as ET

from aipdf import ocr
import mimetypes
import requests


class ToMarkdown:
    """
    Extracts the content of documents of various formats in markdown format.
    """

    def __init__(self):
        """
        Initializes the ToMarkdown class.
        """

    def call(self, file_path_or_url: str, **kwargs) -> str:
        """
        Converts a file to markdown.
        """
        file_extension = self._get_file_extension(file_path_or_url)
        file_content = self._get_file_content(file_path_or_url)

        if file_extension == ".pdf":
            return self._pdf_to_markdown(file_content, **kwargs)

        elif file_extension in (".xml", ".nessus"):
            return self._xml_to_markdown(file_content, **kwargs)

        else:
            raise ValueError(f"Unsupported file type: {file_extension}.")

    def _get_file_content(self, file_path_or_url: str) -> BytesIO:
        """
        Retrieves the content of a file.
        """
        parsed_url = urlparse(file_path_or_url)
        if parsed_url.scheme in ("http", "https"):
            response = requests.get(file_path_or_url)
            if response.status_code == 200:
                return BytesIO(response.content)
            else:
                raise RuntimeError(f"Unable to retrieve file from URL: {file_path_or_url}")
        else:
            with open(file_path_or_url, "rb") as file:
                return BytesIO(file.read())

    def _get_file_extension(self, file_path_or_url: str) -> str:
        """
        Retrieves the file extension from a file path or URL.
        """
        parsed_url = urlparse(file_path_or_url)
        if parsed_url.scheme in ("http", "https"):
            try:
                # Make a HEAD request to get headers without downloading the file.
                response = requests.head(file_path_or_url, allow_redirects=True)
                content_type = response.headers.get("Content-Type", "")
                if content_type:
                    ext = mimetypes.guess_extension(content_type.split(";")[0].strip())
                    if ext:
                        return ext

                # Fallback to extracting extension from the URL path
                ext = os.path.splitext(parsed_url.path)[1]
                if ext:
                    return ext
            except requests.RequestException:
                raise RuntimeError(f"Unable to retrieve file extension from URL: {file_path_or_url}")
        else:
            return os.path.splitext(file_path_or_url)[1]

    def _pdf_to_markdown(self, file_content: Union[requests.Response, BytesIO], **kwargs) -> str:
        """
        Converts a PDF file to markdown.
        """
        markdown_pages = ocr(file_content, **kwargs)
        return "\n\n---\n\n".join(markdown_pages)

    def _xml_to_markdown(self, file_content: Union[requests.Response, BytesIO], **kwargs) -> str:
        """
        Converts an XML (or Nessus) file to markdown.
        """

        def parse_element(element: ET.Element, depth: int = 0) -> str:
            """
            Recursively parses an XML element and converts it to markdown.
            """
            markdown = []
            heading = "#" * (depth + 1)

            markdown.append(f"{heading} {element.tag}")

            for key, val in element.attrib.items():
                markdown.append(f"- **{key}**: {val}")

            text = (element.text or "").strip()
            if text:
                markdown.append(f"\n{text}\n")

            for child in element:
                markdown.append(parse_element(child, depth + 1))

            return "\n".join(markdown)

        root = ET.fromstring(file_content.read().decode("utf-8"))
        markdown_content = parse_element(root)
        return markdown_content



"""Utility functions for RAG pipeline configuration"""
from typing import Dict, Any, Optional

from mindsdb.utilities.log import getLogger
from mindsdb.integrations.utilities.rag.settings import (
    RetrieverType, MultiVectorRetrieverMode, SearchType,
    SearchKwargs, SummarizationConfig, VectorStoreConfig,
    RerankerConfig, RAGPipelineModel, DEFAULT_COLLECTION_NAME
)

logger = getLogger(__name__)


def load_rag_config(base_config: Dict[str, Any], kb_params: Optional[Dict[str, Any]] = None, embedding_model: Any = None) -> RAGPipelineModel:
    """
    Load and validate RAG configuration parameters. This function handles the conversion of configuration
    parameters into their appropriate types and ensures all required settings are properly configured.

    Args:
        base_config: Base configuration dictionary containing RAG pipeline settings
        kb_params: Optional knowledge base parameters to merge with base config
        embedding_model: Optional embedding model instance to use in the RAG pipeline

    Returns:
        RAGPipelineModel: Validated RAG configuration model ready for pipeline creation

    Raises:
        ValueError: If configuration validation fails or required parameters are missing
    """
    # Create a shallow copy of the base config to avoid modifying the original
    # We avoid deepcopy because some objects (like embedding_model) may contain unpickleable objects
    rag_params = base_config.copy()

    # Merge with knowledge base params if provided
    if kb_params:
        rag_params.update(kb_params)

    # Set embedding model if provided
    if embedding_model is not None:
        rag_params['embedding_model'] = embedding_model

    # Handle enums and type conversions
    if 'retriever_type' in rag_params:
        rag_params['retriever_type'] = RetrieverType(rag_params['retriever_type'])
    if 'multi_retriever_mode' in rag_params:
        rag_params['multi_retriever_mode'] = MultiVectorRetrieverMode(rag_params['multi_retriever_mode'])
    if 'search_type' in rag_params:
        rag_params['search_type'] = SearchType(rag_params['search_type'])

    # Handle search kwargs if present
    if 'search_kwargs' in rag_params and isinstance(rag_params['search_kwargs'], dict):
        rag_params['search_kwargs'] = SearchKwargs(**rag_params['search_kwargs'])

    # Handle summarization config if present
    summarization_config = rag_params.get('summarization_config')
    if summarization_config is not None and isinstance(summarization_config, dict):
        rag_params['summarization_config'] = SummarizationConfig(**summarization_config)

    # Handle vector store config
    if 'vector_store_config' in rag_params:
        if isinstance(rag_params['vector_store_config'], dict):
            rag_params['vector_store_config'] = VectorStoreConfig(**rag_params['vector_store_config'])
    else:
        rag_params['vector_store_config'] = {}
        logger.warning(f'No collection_name specified for the retrieval tool, '
                       f"using default collection_name: '{DEFAULT_COLLECTION_NAME}'"
                       f'\nWarning: If this collection does not exist, no data will be retrieved')

    if 'reranker_config' in rag_params:
        rag_params['reranker_config'] = RerankerConfig(**rag_params['reranker_config'])

    # Convert to RAGPipelineModel with validation
    try:
        return RAGPipelineModel(**rag_params)
    except Exception as e:
        logger.error(f"Invalid RAG configuration: {str(e)}")
        raise ValueError(f"Configuration validation failed: {str(e)}")



from typing import Type, List, Any
import re
import json
from pydantic import BaseModel, Field
from langchain_core.tools import BaseTool
from mindsdb_sql_parser.ast import Describe, Select, Identifier, Constant, Star


def llm_str_strip(s):
    length = -1
    while length != len(s):
        length = len(s)

        # remove ```
        if s.startswith("```"):
            s = s[3:]
        if s.endswith("```"):
            s = s[:-3]

        # remove trailing new lines
        s = s.strip("\n")

        # remove extra quotes
        for q in ('"', "'", "`"):
            if s.count(q) == 1:
                s = s.strip(q)
    return s


class KnowledgeBaseListToolInput(BaseModel):
    tool_input: str = Field("", description="An empty string to list all knowledge bases.")


class KnowledgeBaseListTool(BaseTool):
    """Tool for listing knowledge bases in MindsDB."""

    name: str = "kb_list_tool"
    description: str = "List all knowledge bases in MindsDB."
    args_schema: Type[BaseModel] = KnowledgeBaseListToolInput
    db: Any = None

    def _run(self, tool_input: str) -> str:
        """List all knowledge bases."""
        kb_names = self.db.get_usable_knowledge_base_names()
        # Convert list to a formatted string for better readability
        if not kb_names:
            return "No knowledge bases found."
        return json.dumps(kb_names)


class KnowledgeBaseInfoToolInput(BaseModel):
    tool_input: str = Field(
        ...,
        description="A comma-separated list of knowledge base names enclosed between $START$ and $STOP$.",
    )


class KnowledgeBaseInfoTool(BaseTool):
    """Tool for getting information about knowledge bases in MindsDB."""

    name: str = "kb_info_tool"
    description: str = "Get information about knowledge bases in MindsDB."
    args_schema: Type[BaseModel] = KnowledgeBaseInfoToolInput
    db: Any = None

    def _extract_kb_names(self, tool_input: str) -> List[str]:
        """Extract knowledge base names from the tool input."""
        # First, check if the input is already a list (passed directly from include_knowledge_bases)
        if isinstance(tool_input, list):
            return tool_input

        # Next, try to parse it as JSON in case it was serialized as a JSON string
        try:
            parsed_input = json.loads(tool_input)
            if isinstance(parsed_input, list):
                return parsed_input
        except (json.JSONDecodeError, TypeError):
            pass

        # Finally, try the original regex pattern for $START$ and $STOP$ markers
        match = re.search(r"\$START\$(.*?)\$STOP\$", tool_input, re.DOTALL)
        if not match:
            # If no markers found, check if it's a simple comma-separated string
            if "," in tool_input:
                return [kb.strip() for kb in tool_input.split(",")]
            # If it's just a single string without formatting, return it as a single item
            if tool_input.strip():
                return [llm_str_strip(tool_input)]
            return []

        # Extract and clean the knowledge base names
        kb_names_str = match.group(1).strip()
        kb_names = re.findall(r"`([^`]+)`", kb_names_str)

        kb_names = [llm_str_strip(n) for n in kb_names]
        return kb_names

    def _run(self, tool_input: str) -> str:
        """Get information about specified knowledge bases."""
        kb_names = self._extract_kb_names(tool_input)

        if not kb_names:
            return "No valid knowledge base names provided. Please provide knowledge base names as a list, comma-separated string, or enclosed in backticks between $START$ and $STOP$."

        results = []

        for kb_name in kb_names:
            try:
                self.db.check_knowledge_base_permission(Identifier(kb_name))

                # Get knowledge base schema
                schema_result = self.db.run_no_throw(str(Describe(kb_name, type="knowledge_base")))

                if not schema_result:
                    results.append(f"Knowledge base `{kb_name}` not found or has no schema information.")
                    continue

                # Format the results
                kb_info = f"## Knowledge Base: `{kb_name}`\n\n"

                # Schema information
                kb_info += "### Schema Information:\n"
                kb_info += "```\n"

                # Handle different return types for schema_result
                if isinstance(schema_result, str):
                    kb_info += f"{schema_result}\n"
                elif isinstance(schema_result, list):
                    for row in schema_result:
                        if isinstance(row, dict):
                            kb_info += f"{json.dumps(row, indent=2)}\n"
                        else:
                            kb_info += f"{str(row)}\n"
                else:
                    kb_info += f"{str(schema_result)}\n"

                kb_info += "```\n\n"

                # Get sample data
                sample_data = self.db.run_no_throw(
                    str(Select(targets=[Star()], from_table=Identifier(kb_name), limit=Constant(20)))
                )

                # Sample data
                kb_info += "### Sample Data:\n"

                # Handle different return types for sample_data
                if not sample_data:
                    kb_info += "No sample data available.\n"
                elif isinstance(sample_data, str):
                    kb_info += f"```\n{sample_data}\n```\n"
                elif isinstance(sample_data, list) and len(sample_data) > 0:
                    # Only try to extract columns if we have a list of dictionaries
                    if isinstance(sample_data[0], dict):
                        # Extract column names
                        columns = list(sample_data[0].keys())

                        # Create markdown table header
                        kb_info += "| " + " | ".join(columns) + " |\n"
                        kb_info += "| " + " | ".join(["---" for _ in columns]) + " |\n"

                        # Add rows
                        for row in sample_data:
                            formatted_row = []
                            for col in columns:
                                cell_value = row[col]
                                if isinstance(cell_value, dict):
                                    cell_value = json.dumps(cell_value, ensure_ascii=False)
                                formatted_row.append(str(cell_value).replace("|", "\\|"))
                            kb_info += "| " + " | ".join(formatted_row) + " |\n"
                    else:
                        # If it's a list but not of dictionaries, just format as text
                        kb_info += "```\n"
                        for item in sample_data:
                            kb_info += f"{str(item)}\n"
                        kb_info += "```\n"
                else:
                    # For any other type, just convert to string
                    kb_info += f"```\n{str(sample_data)}\n```\n"

                results.append(kb_info)

            except Exception as e:
                results.append(f"Error getting information for knowledge base `{kb_name}`: {str(e)}")

        return "\n\n".join(results)


class KnowledgeBaseQueryToolInput(BaseModel):
    tool_input: str = Field(
        ...,
        description="A SQL query for knowledge bases. Can be provided directly or enclosed between $START$ and $STOP$.",
    )


class KnowledgeBaseQueryTool(BaseTool):
    """Tool for querying knowledge bases in MindsDB."""

    name: str = "kb_query_tool"
    description: str = "Query knowledge bases in MindsDB."
    args_schema: Type[BaseModel] = KnowledgeBaseQueryToolInput
    db: Any = None

    def _extract_query(self, tool_input: str) -> str:
        """Extract the SQL query from the tool input."""
        # First check if the input is wrapped in $START$ and $STOP$
        match = re.search(r"\$START\$(.*?)\$STOP\$", tool_input, re.DOTALL)
        if match:
            return match.group(1).strip()

        # If not wrapped in delimiters, use the input directly
        # Check for SQL keywords to validate it's likely a query
        if re.search(r"\b(SELECT|FROM|WHERE|LIMIT|ORDER BY)\b", tool_input, re.IGNORECASE):
            return tool_input.strip()

        return ""

    def _run(self, tool_input: str) -> str:
        """Execute a knowledge base query."""
        query = self._extract_query(tool_input)

        if not query:
            return "No valid SQL query provided. Please provide a SQL query that includes SELECT, FROM, or other SQL keywords."

        try:
            # Execute the query
            query = llm_str_strip(query)
            result = self.db.run_no_throw(query)

            if not result:
                return "Query executed successfully, but no results were returned."

            # Format the results as a markdown table
            if isinstance(result, list) and len(result) > 0:
                # Extract column names
                columns = list(result[0].keys())

                # Create markdown table header
                table = "| " + " | ".join(columns) + " |\n"
                table += "| " + " | ".join(["---" for _ in columns]) + " |\n"

                # Add rows
                for row in result:
                    formatted_row = []
                    for col in columns:
                        cell_value = row[col]
                        if isinstance(cell_value, dict):
                            cell_value = json.dumps(cell_value, ensure_ascii=False)
                        formatted_row.append(str(cell_value).replace("|", "\\|"))
                    table += "| " + " | ".join(formatted_row) + " |\n"

                return table

            # Ensure we always return a string
            if isinstance(result, (list, dict)):
                return json.dumps(result, indent=2)
            return str(result)
        except Exception as e:
            return f"Error executing query: {str(e)}"



import os
import sys
import json
import argparse
import datetime
import logging
from pathlib import Path
from copy import deepcopy

from appdirs import user_data_dir

# NOTE do not `import from mindsdb` here
logger = logging.getLogger(__name__)


def _merge_key_recursive(target_dict, source_dict, key):
    if key not in target_dict:
        target_dict[key] = source_dict[key]
    elif not isinstance(target_dict[key], dict) or not isinstance(source_dict[key], dict):
        target_dict[key] = source_dict[key]
    else:
        for k in list(source_dict[key].keys()):
            _merge_key_recursive(target_dict[key], source_dict[key], k)


def _merge_configs(original_config: dict, override_config: dict) -> dict:
    for key in list(override_config.keys()):
        _merge_key_recursive(original_config, override_config, key)
    return original_config


def _overwrite_configs(original_config: dict, override_config: dict) -> dict:
    """Overwrite original config with override config."""
    for key in list(override_config.keys()):
        original_config[key] = override_config[key]
    return original_config


def create_data_dir(path: Path) -> None:
    """Create a directory and checks that it is writable.

    Args:
        path (Path): path to create and check

    Raises:
        NotADirectoryError: if path exists, but it is not a directory
        PermissionError: if path exists/created, but it is not writable
        Exception: if directory could not be created
    """
    if path.exists() and not path.is_dir():
        raise NotADirectoryError(f"The path is not a directory: {path}")

    try:
        path.mkdir(mode=0o777, exist_ok=True, parents=True)
    except Exception as e:
        raise Exception("MindsDB storage directory could not be created") from e

    if not os.access(path, os.W_OK):
        raise PermissionError(f"The directory is not allowed for writing: {path}")


class Config:
    """Application config. Singletone, initialized just once. Re-initialyze if `config.auto.json` is changed.
    The class loads multiple configs and merge then in one. If a config option defined in multiple places (config file,
    env var, cmd arg, etc), then it will be resolved in following order of priority:
     - default config values (lowest priority)
     - `config.json` provided by the user
     - `config.auto.json`
     - config values collected from env vars
     - values from cmd args (most priority)

    Attributes:
        __instance (Config): instance of 'Config' to make it singleton
        _config (dict): application config, the result of merging other configs
        _user_config (dict): config provided by the user (usually with cmd arg `--config=config.json`)
        _env_config (dict): config collected from different env vars
        _auto_config (dict): config that is editd by the app itself (e.g. when you change values in GUI)
        _default_config (dict): config with default values
        config_path (Path): path to the `config.json` provided by the user
        storage_root_path (Path): path to storage root folder
        auto_config_path (Path): path to `config.auto.json`
        auto_config_mtime (float): mtime of `config.auto.json` when it was loaded to `self._auto_config`
        _cmd_args (argparse.Namespace): cmd args
        use_docker_env (bool): is the app run in docker env
    """

    __instance: "Config" = None

    _config: dict = None
    _user_config: dict = None
    _env_config: dict = None
    _auto_config: dict = None
    _default_config: dict = None
    config_path: Path = None
    storage_root_path: Path = None
    auto_config_path: Path = None
    auto_config_mtime: float = 0
    _cmd_args: argparse.Namespace = None
    use_docker_env: bool = os.environ.get("MINDSDB_DOCKER_ENV", False) is not False

    def __new__(cls, *args, **kwargs) -> "Config":
        """Make class singletone and initialize config."""
        if cls.__instance is not None:
            return cls.__instance

        self = super().__new__(cls, *args, **kwargs)
        cls.__instance = self

        self.fetch_user_config()

        # region determine root path
        if self.storage_root_path is None:
            if isinstance(os.environ.get("MINDSDB_STORAGE_DIR"), str):
                self.storage_root_path = os.environ["MINDSDB_STORAGE_DIR"]
            elif "root" in self._user_config.get("paths", {}):
                self.storage_root_path = self.user_config["paths"]["root"]
            else:
                self.storage_root_path = os.path.join(user_data_dir("mindsdb", "mindsdb"), "var/")
            self.storage_root_path = Path(self.storage_root_path)
            create_data_dir(self.storage_root_path)
        # endregion

        # region prepare default config
        api_host = "127.0.0.1" if not self.use_docker_env else "0.0.0.0"
        self._default_config = {
            "permanent_storage": {"location": "absent"},
            "storage_db": (
                "sqlite:///"
                + str(self.storage_root_path / "mindsdb.sqlite3.db")
                + "?check_same_thread=False&timeout=30"
            ),
            "paths": {
                "root": self.storage_root_path,
                "content": self.storage_root_path / "content",
                "storage": self.storage_root_path / "storage",
                "static": self.storage_root_path / "static",
                "tmp": self.storage_root_path / "tmp",
                "log": self.storage_root_path / "log",
                "cache": self.storage_root_path / "cache",
                "locks": self.storage_root_path / "locks",
            },
            "auth": {
                "http_auth_enabled": False,
                "http_permanent_session_lifetime": datetime.timedelta(days=31),
                "username": "mindsdb",
                "password": "",
            },
            "logging": {
                "handlers": {
                    "console": {
                        "enabled": True,
                        "formatter": "default",
                        "level": "INFO",  # MINDSDB_CONSOLE_LOG_LEVEL or MINDSDB_LOG_LEVEL (obsolete)
                    },
                    "file": {
                        "enabled": False,
                        "level": "INFO",  # MINDSDB_FILE_LOG_LEVEL
                        "filename": "app.log",
                        "maxBytes": 1 << 19,  # 0.5 Mb
                        "backupCount": 3,
                    },
                }
            },
            "gui": {"autoupdate": True},
            "debug": False,
            "environment": "local",
            "integrations": {},
            "api": {
                "http": {
                    "host": api_host,
                    "port": "47334",
                    "restart_on_failure": True,
                    "max_restart_count": 1,
                    "max_restart_interval_seconds": 60,
                },
                "mysql": {
                    "host": api_host,
                    "port": "47335",
                    "database": "mindsdb",
                    "ssl": True,
                    "restart_on_failure": True,
                    "max_restart_count": 1,
                    "max_restart_interval_seconds": 60,
                },
                "postgres": {"host": api_host, "port": "55432", "database": "mindsdb"},
                "litellm": {
                    "host": "0.0.0.0",  # API server binds to all interfaces by default
                    "port": "8000",
                },
            },
            "cache": {"type": "local"},
            "ml_task_queue": {"type": "local"},
            "url_file_upload": {"enabled": True, "allowed_origins": [], "disallowed_origins": []},
            "file_upload_domains": [],  # deprecated, use config[url_file_upload][allowed_origins] instead
            "web_crawling_allowed_sites": [],
            "cloud": False,
            "jobs": {"disable": False},
            "tasks": {"disable": False},
            "default_project": "mindsdb",
            "default_llm": {},
            "default_embedding_model": {},
            "default_reranking_model": {},
            "data_catalog": {
                "enabled": False,
            },
        }
        # endregion

        # region find 'auto' config file, create if not exists
        auto_config_name = "config.auto.json"
        auto_config_path = self.storage_root_path.joinpath(auto_config_name)
        if not auto_config_path.is_file():
            auto_config_path.write_text("{}")
        self.auto_config_path = auto_config_path
        # endregion

        self.prepare_env_config()

        self.fetch_auto_config()
        self.merge_configs()

        return cls.__instance

    def prepare_env_config(self) -> None:
        """Collect config values from env vars to self._env_config"""
        self._env_config = {
            "logging": {"handlers": {"console": {}, "file": {}}},
            "api": {"http": {}},
            "auth": {},
            "paths": {},
            "permanent_storage": {},
            "ml_task_queue": {},
        }

        # region storage root path
        if os.environ.get("MINDSDB_STORAGE_DIR", "") != "":
            self._env_config["paths"] = {"root": Path(os.environ["MINDSDB_STORAGE_DIR"])}
        # endregion

        # region vars: permanent storage disabled?
        if os.environ.get("MINDSDB_STORAGE_BACKUP_DISABLED", "").lower() in (
            "1",
            "true",
        ):
            self._env_config["permanent_storage"] = {"location": "absent"}
        # endregion

        # region vars: ml queue
        if os.environ.get("MINDSDB_ML_QUEUE_TYPE", "").lower() == "redis":
            self._env_config["ml_task_queue"] = {
                "type": "redis",
                "host": os.environ.get("MINDSDB_ML_QUEUE_HOST", "localhost"),
                "port": int(os.environ.get("MINDSDB_ML_QUEUE_PORT", 6379)),
                "db": int(os.environ.get("MINDSDB_ML_QUEUE_DB", 0)),
                "username": os.environ.get("MINDSDB_ML_QUEUE_USERNAME"),
                "password": os.environ.get("MINDSDB_ML_QUEUE_PASSWORD"),
            }
        # endregion

        # region vars: username and password
        http_username = os.environ.get("MINDSDB_USERNAME")
        http_password = os.environ.get("MINDSDB_PASSWORD")

        if bool(http_username) != bool(http_password):
            raise ValueError(
                "Both MINDSDB_USERNAME and MINDSDB_PASSWORD must be set together and must be non-empty strings."
            )

        # If both username and password are set, enable HTTP auth.
        if http_username and http_password:
            self._env_config["auth"]["http_auth_enabled"] = True
            self._env_config["auth"]["username"] = http_username
            self._env_config["auth"]["password"] = http_password
        # endregion

        # region logging
        if os.environ.get("MINDSDB_LOG_LEVEL", "") != "":
            self._env_config["logging"]["handlers"]["console"]["level"] = os.environ["MINDSDB_LOG_LEVEL"]
            self._env_config["logging"]["handlers"]["console"]["enabled"] = True
        if os.environ.get("MINDSDB_CONSOLE_LOG_LEVEL", "") != "":
            self._env_config["logging"]["handlers"]["console"]["level"] = os.environ["MINDSDB_CONSOLE_LOG_LEVEL"]
            self._env_config["logging"]["handlers"]["console"]["enabled"] = True
        if os.environ.get("MINDSDB_FILE_LOG_LEVEL", "") != "":
            self._env_config["logging"]["handlers"]["file"]["level"] = os.environ["MINDSDB_FILE_LOG_LEVEL"]
            self._env_config["logging"]["handlers"]["file"]["enabled"] = True
        # endregion

        if os.environ.get("MINDSDB_DB_CON", "") != "":
            self._env_config["storage_db"] = os.environ["MINDSDB_DB_CON"]

        if os.environ.get("MINDSDB_DEFAULT_PROJECT", "") != "":
            self._env_config["default_project"] = os.environ["MINDSDB_DEFAULT_PROJECT"].lower()

        if os.environ.get("MINDSDB_DEFAULT_LLM_API_KEY", "") != "":
            self._env_config["default_llm"] = {"api_key": os.environ["MINDSDB_DEFAULT_LLM_API_KEY"]}
        if os.environ.get("MINDSDB_DEFAULT_EMBEDDING_MODEL_API_KEY", "") != "":
            self._env_config["default_embedding_model"] = {
                "api_key": os.environ["MINDSDB_DEFAULT_EMBEDDING_MODEL_API_KEY"]
            }
        if os.environ.get("MINDSDB_DEFAULT_RERANKING_MODEL_API_KEY", "") != "":
            self._env_config["default_reranking_model"] = {
                "api_key": os.environ["MINDSDB_DEFAULT_RERANKING_MODEL_API_KEY"]
            }
        if os.environ.get("MINDSDB_DATA_CATALOG_ENABLED", "").lower() in ("1", "true"):
            self._env_config["data_catalog"] = {"enabled": True}

    def fetch_auto_config(self) -> bool:
        """Load dict readed from config.auto.json to `auto_config`.
        Do it only if `auto_config` was not loaded before or config.auto.json been changed.

        Returns:
            bool: True if config was loaded or updated
        """

        if (
            self.auto_config_path.is_file()
            and self.auto_config_path.read_text() != ""
            and self.auto_config_mtime != self.auto_config_path.stat().st_mtime
        ):
            try:
                self._auto_config = json.loads(self.auto_config_path.read_text())
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"The 'auto' configuration file ({self.auto_config_path}) contains invalid JSON: {e}\nFile content: {self.auto_config_path.read_text()}"
                )
            self.auto_config_mtime = self.auto_config_path.stat().st_mtime
            return True
        return False

    def fetch_user_config(self) -> bool:
        """Read config provided by the user to `user_config`. Do it only if `user_config` was not loaded before.

        Returns:
            bool: True if config was loaded
        """
        if self._user_config is None:
            cmd_args_config = self.cmd_args.config
            if isinstance(cmd_args_config, str):
                self.config_path = cmd_args_config
            elif isinstance(os.environ.get("MINDSDB_CONFIG_PATH"), str):
                self.config_path = os.environ["MINDSDB_CONFIG_PATH"]
            if self.config_path == "absent":
                self.config_path = None
            if isinstance(self.config_path, str):
                self.config_path = Path(self.config_path)
                if not self.config_path.is_file():
                    raise FileNotFoundError(f"The configuration file was not found at the path: {self.config_path}")
                try:
                    self._user_config = json.loads(self.config_path.read_text())
                except json.JSONDecodeError as e:
                    raise ValueError(f"The configuration file ({self.config_path}) contains invalid JSON: {e}")
            else:
                self._user_config = {}
            return True
        return False

    def ensure_auto_config_is_relevant(self) -> None:
        """Check if auto config has not been changed. If changed - reload main config."""
        updated = self.fetch_auto_config()
        if updated:
            self.merge_configs()

    def merge_configs(self) -> None:
        """Merge multiple configs to one."""
        new_config = deepcopy(self._default_config)
        _merge_configs(new_config, self._user_config)
        _merge_configs(new_config, self._auto_config or {})
        _merge_configs(new_config, self._env_config or {})

        # region create dirs
        for key, value in new_config["paths"].items():
            if isinstance(value, str):
                new_config["paths"][key] = Path(value)
            elif isinstance(value, Path) is False:
                raise ValueError(f"Unexpected path value: {value}")
            create_data_dir(new_config["paths"][key])
        # endregion

        self._config = new_config

    def __getitem__(self, key):
        self.ensure_auto_config_is_relevant()
        return self._config[key]

    def get(self, key, default=None):
        self.ensure_auto_config_is_relevant()
        return self._config.get(key, default)

    def get_all(self):
        self.ensure_auto_config_is_relevant()
        return self._config

    def update(self, data: dict, overwrite: bool = False) -> None:
        """
        Update values in `auto` config.
        Args:
            data (dict): data to update in `auto` config.
            overwrite (bool): if True, overwrite existing keys, otherwise merge them.
                - False (default): Merge recursively. Existing nested dictionaries are preserved
                and only the specified keys in `data` are updated.
                - True: Overwrite completely. Existing keys are replaced entirely with values
                from `data`, discarding any nested structure not present in `data`.
        """
        self.ensure_auto_config_is_relevant()

        if overwrite:
            _overwrite_configs(self._auto_config, data)
        else:
            _merge_configs(self._auto_config, data)

        self.auto_config_path.write_text(json.dumps(self._auto_config, indent=4))

        self.auto_config_mtime = self.auto_config_path.stat().st_mtime

        self.merge_configs()

    def raise_warnings(self, logger) -> None:
        """Show warnings about config options"""

        if "storage_dir" in self._config:
            logger.warning("The 'storage_dir' config option is no longer supported. Use 'paths.root' instead.")

        if "log" in self._config:
            logger.warning("The 'log' config option is no longer supported. Use 'logging' instead.")

        file_upload_domains = self._config.get("file_upload_domains")
        if isinstance(file_upload_domains, list) and len(file_upload_domains) > 0:
            allowed_origins = self._config["url_file_upload"]["allowed_origins"]
            if isinstance(allowed_origins, list) and len(allowed_origins) == 0:
                self._config["url_file_upload"]["allowed_origins"] = file_upload_domains
            logger.warning(
                'Config option "file_upload_domains" is deprecated, '
                'use config["url_file_upload"]["allowed_origins"] instead.'
            )

    @property
    def cmd_args(self):
        if self._cmd_args is None:
            self.parse_cmd_args()
        return self._cmd_args

    def parse_cmd_args(self) -> None:
        """Collect cmd args to self._cmd_args (accessable as self.cmd_args)"""
        if self._cmd_args is not None:
            return

        # if it is not mindsdb run, then set args to empty
        if (sys.modules["__main__"].__package__ or "").lower() != "mindsdb" and os.environ.get(
            "MINDSDB_RUNTIME"
        ) != "1":
            self._cmd_args = argparse.Namespace(
                api=None,
                config=None,
                install_handlers=None,
                verbose=False,
                no_studio=False,
                version=False,
                ml_task_queue_consumer=None,
                agent=None,
                project=None,
                update_gui=False,
            )
            return

        parser = argparse.ArgumentParser(description="CL argument for mindsdb server")
        parser.add_argument("--api", type=str, default=None)
        parser.add_argument("--config", type=str, default=None)
        parser.add_argument("--install-handlers", type=str, default=None)
        parser.add_argument("--verbose", action="store_true")
        parser.add_argument("--no_studio", action="store_true")
        parser.add_argument("-v", "--version", action="store_true")
        parser.add_argument("--ml_task_queue_consumer", action="store_true", default=None)
        parser.add_argument(
            "--agent",
            type=str,
            default=None,
            help="Name of the agent to use with litellm APIs",
        )
        parser.add_argument(
            "--project",
            type=str,
            default=None,
            help="Project containing the agent (default: mindsdb)",
        )

        parser.add_argument("--project-name", type=str, default=None, help="MindsDB project name")
        parser.add_argument("--update-gui", action="store_true", default=False, help="Update GUI and exit")

        self._cmd_args = parser.parse_args()

    @property
    def paths(self):
        return self._config["paths"]

    @property
    def user_config(self):
        return self._user_config

    @property
    def auto_config(self):
        return self._auto_config

    @property
    def env_config(self):
        return self._env_config

    @property
    def is_cloud(self):
        return self._config.get("cloud", False)


config = Config()



import os.path
import json
import pandas as pd
import pytz
from tzlocal import get_localzone
from datetime import datetime

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from mindsdb_sql_parser import parse_sql

from mindsdb.utilities import log
from mindsdb.integrations.handlers.google_fit_handler.google_fit_tables import GoogleFitTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)

SCOPES = ["https://www.googleapis.com/auth/fitness.activity.read"]
DATE_FORMAT = "%Y-%m-%d"

logger = log.getLogger(__name__)


class GoogleFitHandler(APIHandler):
    def __init__(self, name: str = None, **kwargs):
        super().__init__(name)
        args = kwargs.get("connection_data", {})
        self.connection_args = {}
        self.credentials_path = None
        if "service_account_file" in args:
            if os.path.isfile(args["service_account_file"]) is False:
                raise Exception("service_account_file must be a path to the credentials.json file")
            self.credentials_path = args["service_account_file"]
        elif "service_account_json" in args:
            self.connection_args = args["service_account_json"]
            if (
                not isinstance(self.connection_args, dict)
                or (("redirect_uris" not in self.connection_args.keys()) and len(self.connection_args) != 6)
                or ("redirect_uris" in self.connection_args.keys())
                and len(self.connection_args) != 7
            ):
                raise Exception("service_account_json has to be a dictionary with all 6 required fields")
            self.connection_args["redirect_uris"] = ["http://localhost"]
            self.credentials_path = "mindsdb/integrations/handlers/google_fit_handler/credentials.json"
        else:
            raise Exception("Connection args have to content ether service_account_file or service_account_json")

        self.api = None
        self.is_connected = False

        aggregated_data = GoogleFitTable(self)
        self._register_table("aggregated_data", aggregated_data)

    def connect(self) -> Resource:
        if self.is_connected is True and self.api:
            return self.api
        if self.connection_args:
            credentialDict = {"installed": self.connection_args}
            f = open(self.credentials_path, "w")
            f.write(json.dumps(credentialDict).replace(" ", ""))
            f.close()

        creds = None

        if os.path.isfile("mindsdb/integrations/handlers/google_fit_handler/token.json"):
            creds = Credentials.from_authorized_user_file(
                "mindsdb/integrations/handlers/google_fit_handler/token.json", SCOPES
            )
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(self.credentials_path, SCOPES)
                creds = flow.run_local_server(port=0)
            with open("mindsdb/integrations/handlers/google_fit_handler/token.json", "w") as token:
                token.write(creds.to_json())
        self.api = build("fitness", "v1", credentials=creds)

        self.is_connected = True
        return self.api

    def check_connection(self) -> StatusResponse:
        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True

        except Exception as e:
            logger.error(f"Error connecting to Google Fit API: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def retrieve_data(self, service, startTimeMillis, endTimeMillis, dataSourceId) -> dict:
        try:
            return (
                service.users()
                .dataset()
                .aggregate(
                    userId="me",
                    body={
                        "aggregateBy": [{"dataTypeName": "com.google.step_count.delta", "dataSourceId": dataSourceId}],
                        "bucketByTime": {"durationMillis": 86400000},
                        "startTimeMillis": startTimeMillis,
                        "endTimeMillis": endTimeMillis,
                    },
                )
                .execute()
            )
        except HttpError:
            raise HttpError

    def native_query(self, query: str = None) -> Response:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
            api's json etc)
        Returns:
            HandlerResponse
        """
        ast = parse_sql(query)
        return self.query(ast)

    def get_steps(self, start_time_millis, end_time_millis) -> pd.DataFrame:
        steps = {}
        steps_data = self.retrieve_data(
            self.api,
            start_time_millis,
            end_time_millis,
            "derived:com.google.step_count.delta:com.google.android.gms:estimated_steps",
        )
        for daily_step_data in steps_data["bucket"]:
            local_date = datetime.fromtimestamp(
                int(daily_step_data["startTimeMillis"]) / 1000, tz=pytz.timezone(str(get_localzone()))
            )
            local_date_str = local_date.strftime(DATE_FORMAT)

            data_point = daily_step_data["dataset"][0]["point"]
            if data_point:
                count = data_point[0]["value"][0]["intVal"]
                data_source_id = data_point[0]["originDataSourceId"]
                steps[local_date_str] = {"steps": count, "originDataSourceId": data_source_id}
        ret = pd.DataFrame.from_dict(steps)
        ret = ret.T
        ret = ret.drop("originDataSourceId", axis=1)
        ret = ret.reset_index(drop=False)
        return ret

    def call_google_fit_api(self, method_name: str = None, params: dict = None) -> pd.DataFrame:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc
        Returns:
            DataFrame
        """
        self.connect()
        if method_name == "get_steps":
            val = self.get_steps(params["start_time"], params["end_time"])
            return val
        raise NotImplementedError("Method name {} not supported by Google Fit Handler".format(method_name))



from mindsdb.api.executor.sql_query import SQLQuery
from mindsdb.interfaces.query_context.context_controller import query_context_controller
from mindsdb.api.executor.controllers.session_controller import SessionController
from mindsdb.interfaces.tasks.task import BaseTask


class QueryTask(BaseTask):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.query_id = self.object_id

    def run(self, stop_event):

        try:
            session = SessionController()
            SQLQuery(None, query_id=self.query_id, session=session, stop_event=stop_event)
        finally:
            # clear task
            query_context_controller.get_query(self.query_id).remove_from_task()



import unittest
from mindsdb.integrations.handlers.tidb_handler.tidb_handler import TiDBHandler


class TiDBHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "host": "localhost",
            "port": "4000",
            "user": "root",
            "password": "root",
            "database": "test",
            "ssl": False
        }
        cls.handler = TiDBHandler('test_tidb_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()

    def test_1_check_connection(self):
        assert self.handler.check_connection()

    def test_2_native_query_show_dbs(self):
        dbs = self.handler.native_query("SHOW DATABASES;")
        assert isinstance(dbs, list)

    def test_3_get_tables(self):
        tbls = self.handler.get_tables()
        assert isinstance(tbls, list)

    def test_5_create_table(self):
        try:
            self.handler.native_query("CREATE TABLE test_tidb (test_col INT)")
        except Exception:
            pass

    def test_6_describe_table(self):
        described = self.handler.get_columns("dt_test")
        assert isinstance(described, list)

    def test_7_select_query(self):
        query = "SELECT * FROM dt_test WHERE 'id'='a'"
        self.handler.native_query(query)



"""added_encrypted_content_to_json_storage

Revision ID: 4521dafe89ab
Revises: 6ab9903fc59a
Create Date: 2025-02-14 12:05:13.102594

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db  # noqa


# revision identifiers, used by Alembic.
revision = '4521dafe89ab'
down_revision = '6ab9903fc59a'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('json_storage', schema=None) as batch_op:
        batch_op.add_column(sa.Column('encrypted_content', sa.LargeBinary(), nullable=True))
        batch_op.alter_column('resource_id', existing_type=sa.Integer(), type_=sa.BigInteger())


def downgrade():
    with op.batch_alter_table('json_storage', schema=None) as batch_op:
        batch_op.drop_column('encrypted_content')
        batch_op.alter_column('resource_id', existing_type=sa.BigInteger(), type_=sa.Integer())



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


creation_args = OrderedDict(
    togetherai_api_key={
        "type": ARG_TYPE.STR,
        "description": "Key for TogetherAI API.",
        "required": False,
        "label": "TogetherAI API key",
        "secret": True,
    }
)



from flask import request
from flask_restx import Resource
from flask_restx import fields

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.namespaces.configs.default import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities.config import Config
from mindsdb.utilities import log
from mindsdb.api.common.middleware import generate_pat, revoke_pat, verify_pat


logger = log.getLogger(__name__)


@ns_conf.route("/login", methods=["POST"])
class LoginRoute(Resource):
    @ns_conf.doc(
        responses={200: "Success", 400: "Error in username or password", 401: "Invalid username or password"},
        body=ns_conf.model(
            "request_login",
            {"username": fields.String(description="Username"), "password": fields.String(description="Password")},
        ),
    )
    @api_endpoint_metrics("POST", "/default/login")
    def post(self):
        """Check user's credentials and creates a session"""
        username = request.json.get("username")
        password = request.json.get("password")
        if (
            isinstance(username, str) is False
            or len(username) == 0
            or isinstance(password, str) is False
            or len(password) == 0
        ):
            return http_error(400, "Error in username or password", "Username and password should be string")

        config = Config()
        inline_username = config["auth"]["username"]
        inline_password = config["auth"]["password"]

        if username != inline_username or password != inline_password:
            return http_error(401, "Forbidden", "Invalid username or password")

        logger.info(f"User '{username}' logged in successfully")

        return {"token": generate_pat()}, 200


@ns_conf.route("/logout", methods=["POST"])
class LogoutRoute(Resource):
    @ns_conf.doc(responses={200: "Success"})
    @api_endpoint_metrics("POST", "/default/logout")
    def post(self):
        # We can't forcibly log out a user with the
        h = request.headers.get("Authorization")
        if not h or not h.startswith("Bearer "):
            bearer = None
        else:
            bearer = h.split(" ", 1)[1].strip() or None
        revoke_pat(bearer)
        return "", 200


@ns_conf.route("/status")
class StatusRoute(Resource):
    @ns_conf.doc(
        responses={200: "Success"},
        model=ns_conf.model(
            "response_status",
            {
                "environment": fields.String(description="The name of current environment: cloud, local or other"),
                "mindsdb_version": fields.String(description="Current version of mindsdb"),
                "auth": fields.Nested(
                    ns_conf.model(
                        "response_status_auth",
                        {
                            "confirmed": fields.Boolean(description="is current user authenticated"),
                            "required": fields.Boolean(description="is authenticated required"),
                            "provider": fields.Boolean(description="current authenticated provider: local of 3d-party"),
                        },
                    )
                ),
            },
        ),
    )
    @api_endpoint_metrics("GET", "/default/status")
    def get(self):
        """returns auth and environment data"""
        environment = "local"
        config = Config()

        environment = config.get("environment")
        if environment is None:
            if config.get("cloud", False):
                environment = "cloud"
            elif config.get("aws_marketplace", False):
                environment = "aws_marketplace"
            else:
                environment = "local"

        auth_provider = "disabled"
        if config["auth"]["http_auth_enabled"] is True:
            if config["auth"].get("provider") is not None:
                auth_provider = config["auth"].get("provider")
            else:
                auth_provider = "local"

        resp = {
            "mindsdb_version": mindsdb_version,
            "environment": environment,
            "auth": {
                "confirmed": verify_pat(request.headers.get("Authorization", "").replace("Bearer ", "")),
                "http_auth_enabled": config["auth"]["http_auth_enabled"],
                "provider": auth_provider,
            },
        }

        return resp



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    api_base={
        "type": ARG_TYPE.URL,
        "description": "The base URL of the Confluence instance/server.",
        "label": "Base URL",
        "required": True
    },
    username={
        "type": ARG_TYPE.STR,
        "description": "The username for the Confluence account.",
        "label": "Username",
        "required": True
    },
    password={
        "type": ARG_TYPE.STR,
        "description": "The API token for the Confluence account.",
        "label": "Password",
        "required": True,
        "secret": True
    }
)

connection_args_example = OrderedDict(
    api_base="https://marios.atlassian.net/",
    username="your_username",
    password="access_token"
)



import pandas as pd
from ckanapi import RemoteCKAN

from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb.integrations.libs.response import (
    HandlerResponse,
    HandlerStatusResponse,
    RESPONSE_TYPE,
)
from mindsdb_sql_parser import ast
from mindsdb.utilities import log
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions

logger = log.getLogger(__name__)


class DatasetsTable(APITable):
    '''
    Datasets table contains information about CKAN datasets.
    This table is used to list all datasets available in CKAN that have datastore active resources.
    '''

    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where) if query.where else []
        limit = query.limit.value if query.limit else 1000
        packages = self.list(conditions, limit)
        return pd.DataFrame(packages)

    def list(self, conditions=None, limit=1000):
        self.handler.connect()
        package_list = self.handler.call_ckan_api("package_search", {"rows": limit})
        packages = package_list.get("results", [])

        data = []
        # Get only datastore active resources
        for pkg in packages:
            datastore_active_resources = [
                r for r in pkg.get("resources", []) if r.get("datastore_active")
            ]
            data.append(
                {
                    "id": pkg.get("id"),
                    "name": pkg.get("name"),
                    "title": pkg.get("title"),
                    "num_resources": len(pkg.get("resources", [])),
                    "num_datastore_active_resources": len(datastore_active_resources),
                }
            )

        return pd.DataFrame(data)

    def get_columns(self):
        return [
            "id",
            "name",
            "title",
            "num_resources",
            "num_datastore_active_resources",
        ]


class ResourceIDsTable(APITable):
    '''
    ResourceIDs table contains information about CKAN resources.
    This table is used to list all resources available in CKAN that are datastore active.
    '''

    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where) if query.where else []
        limit = query.limit.value if query.limit else 1000

        resources = self.list(conditions, limit)
        return pd.DataFrame(resources)

    def list(self, conditions=None, limit=1000):
        self.handler.connect()
        package_list = self.handler.call_ckan_api("package_search", {"rows": limit})
        packages = package_list.get("results", [])

        data = []
        for package in packages:
            for resource in package.get("resources", []):
                # Get only datastore active resources
                if resource.get("datastore_active"):
                    data.append(
                        {
                            "id": resource.get("id"),
                            "package_id": package.get("id"),
                            "name": resource.get("name"),
                            "format": resource.get("format"),
                            "url": resource.get("url"),
                            "datastore_active": resource.get("datastore_active"),
                        }
                    )
                if len(data) >= limit:
                    break
            if len(data) >= limit:
                break

        return pd.DataFrame(data)

    def get_columns(self):
        return [
            "id",
            "package_id",
            "name",
            "format",
            "url",
            "datastore_active",
        ]


class DatastoreTable(APITable):
    '''
    Datastore table is used to query CKAN datastore resources.
    This table is used to query data from CKAN datastore resources.
    It is using the datastore_search_sql API to execute SQL queries on CKAN datastore resources.
    '''

    def select(self, query: ast.Select) -> pd.DataFrame:
        conditions = extract_comparison_conditions(query.where) if query.where else []
        resource_id = self.extract_resource_id(conditions)

        if resource_id:
            return self.execute_resource_query(query, resource_id)
        else:
            message = "Please provide a resource_id in your query. Example: SELECT * FROM datastore WHERE resource_id = 'your_resource_id'"
            df = pd.DataFrame({"message": [message]})
            return df

    def execute_resource_query(
        self, query: ast.Select, resource_id: str
    ) -> pd.DataFrame:
        sql_query = self.ast_to_sql(query, resource_id)
        result = self.handler.call_ckan_api("datastore_search_sql", {"sql": sql_query})

        records = result.get("records", [])

        df = pd.DataFrame(records)

        df = df.loc[:, ~df.columns.str.startswith("_")]

        return df

    def ast_to_sql(self, query: ast.Select, resource_id: str) -> str:
        sql_parts = [
            f"SELECT {self.render_columns(query.targets)}",
            f'FROM "{resource_id}"',
        ]

        # Handle WHERE clause
        where_conditions = []
        if query.where:
            where_conditions = self.extract_where_conditions(query.where)

        where_conditions = [
            cond
            for cond in where_conditions
            if not (
                isinstance(cond, ast.BinaryOperation)
                and cond.args[0].parts[-1] == "resource_id"
            )
        ]

        if where_conditions:
            sql_parts.append(
                f'WHERE {" AND ".join(self.render_where(cond) for cond in where_conditions)}'
            )

        # Handle LIMIT
        if query.limit:
            sql_parts.append(f"LIMIT {query.limit.value}")

        return " ".join(sql_parts)

    def render_columns(self, targets):
        if not targets or (len(targets) == 1 and isinstance(targets[0], ast.Star)):
            return "*"
        return ", ".join(self.render_column(target) for target in targets)

    def render_column(self, target):
        if isinstance(target, ast.Identifier):
            return f'"{target.parts[-1]}"'
        # Handle other types of targets as needed
        return str(target)

    def extract_where_conditions(self, where):
        if isinstance(where, ast.BinaryOperation) and where.op == "and":
            return self.extract_where_conditions(
                where.args[0]
            ) + self.extract_where_conditions(where.args[1])
        return [where]

    def render_where(self, where):
        if isinstance(where, ast.BinaryOperation):
            left = self.render_where(where.args[0])
            right = self.render_where(where.args[1])

            if where.op == "like":
                return f"{left} ILIKE {right}"
            elif where.op in ["=", ">", "<", ">=", "<=", "<>"]:
                return f"{left} {where.op} {right}"
            # Add more operators as needed

        elif isinstance(where, ast.Constant):
            return (
                f"'{where.value}'" if isinstance(where.value, str) else str(where.value)
            )

        elif isinstance(where, ast.Identifier):
            return f'"{where.parts[-1]}"'

        # Handle other types of WHERE conditions as needed
        return str(where)

    def extract_resource_id(self, conditions):
        for condition in conditions:
            if isinstance(condition, list) and len(condition) == 3:
                op, col, val = condition
                if col == "resource_id" and op == "=":
                    return val
        return None

    def get_columns(self):
        return [field["id"] for field in self.fields]


class CkanHandler(APIHandler):
    name = "ckan"

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.connection = None
        self.is_connected = False
        self.connection_args = kwargs.get("connection_data", {})

        self.datasets_table = DatasetsTable(self)
        self.resources_table = ResourceIDsTable(self)
        self.datastore_table = DatastoreTable(self)

        self._register_table("datasets", self.datasets_table)
        self._register_table("resources", self.resources_table)
        self._register_table("datastore", self.datastore_table)

    def connect(self):
        if self.is_connected:
            return self.connection

        url = self.connection_args.get("url")
        api_key = self.connection_args.get("api_key")
        if not url:
            raise ValueError("CKAN URL is required")

        try:
            self.connection = RemoteCKAN(url, apikey=api_key)
            self.is_connected = True
            logger.info(f"Successfully connected to CKAN at {url}")
        except Exception as e:
            logger.error(f"Error connecting to CKAN: {e}")
            raise ConnectionError(f"Failed to connect to CKAN: {e}")

        return self.connection

    def check_connection(self) -> HandlerStatusResponse:
        try:
            self.connect()
            return HandlerStatusResponse(success=True)
        except Exception as e:
            logger.error(f"Error checking connection: {e}")
            return HandlerStatusResponse(success=False, error_message=str(e))

    def call_ckan_api(self, method_name: str, params: dict):
        connection = self.connect()
        method = getattr(connection.action, method_name)

        try:
            result = method(**params)
            return result
        except Exception as e:
            logger.error(f"Error calling CKAN API: {e}")
            raise RuntimeError(f"Failed to call CKAN API: {e}")

    def native_query(self, query: str) -> HandlerResponse:
        method, params = self.parse_native_query(query)
        try:
            result = self.call_ckan_api(method, params)
            if isinstance(result, list):
                df = pd.DataFrame(result)
            elif isinstance(result, dict):
                df = pd.DataFrame([result])
            else:
                df = pd.DataFrame([{"result": result}])
            return HandlerResponse(RESPONSE_TYPE.TABLE, df)
        except Exception as e:
            logger.error(f"Error executing native query: {e}")
            return HandlerResponse(RESPONSE_TYPE.ERROR, error_message=str(e))

    @staticmethod
    def parse_native_query(query: str):
        parts = query.split(":")
        if len(parts) != 2:
            raise ValueError(
                "Invalid query format. Expected 'method_name:param1=value1,param2=value2'"
            )
        method = parts[0].strip()
        params = {}
        if parts[1].strip():
            param_pairs = parts[1].split(",")
            for pair in param_pairs:
                key, value = pair.split("=")
                params[key.strip()] = value.strip()

        return method, params



from collections.abc import Callable

import redis
from pandas import DataFrame

from mindsdb.utilities.ml_task_queue.utils import RedisKey, from_bytes
from mindsdb.utilities.ml_task_queue.const import ML_TASK_STATUS


class Task:
    """Abstraction for ML task. Should have interface similat to concurrent.futures.Future

    Attributes:
        db (Redis): database object
        redis_key (RedisKey): redis keys associated with task
        dataframe (DataFrame): task result
        exception (Exception): task exeuton  runtime exception
        _timeout (int): max time without status updating
    """

    def __init__(self, connection: redis.Redis, redis_key: RedisKey) -> None:
        self.db = connection
        self.redis_key = redis_key
        self.dataframe = None
        self.exception = None
        self._timeout = 60

    def subscribe(self) -> ML_TASK_STATUS:
        """return tasks status untill it is not done or failed"""
        pubsub = self.db.pubsub()
        cache = self.db.cache()
        pubsub.subscribe(self.redis_key.status)
        while msg := pubsub.get_message(timeout=self._timeout):
            if msg["type"] not in pubsub.PUBLISH_MESSAGE_TYPES:
                continue
            ml_task_status = ML_TASK_STATUS(msg["data"])
            if ml_task_status == ML_TASK_STATUS.COMPLETE:
                dataframe_bytes = cache.get(self.redis_key.dataframe)
                if dataframe_bytes is not None:
                    self.dataframe = from_bytes(dataframe_bytes)
                cache.delete(self.redis_key.dataframe)
            elif ml_task_status == ML_TASK_STATUS.ERROR:
                exception_bytes = cache.get(self.redis_key.exception)
                if exception_bytes is not None:
                    self.exception = from_bytes(exception_bytes)
            yield ml_task_status
        else:
            # there is no mesasges, timeout
            ml_task_status = ML_TASK_STATUS.TIMEOUT
            yield ml_task_status

    def wait(self, status: ML_TASK_STATUS = ML_TASK_STATUS.COMPLETE) -> None:
        """block threasd untill task is not done or failed"""
        for status in self.subscribe():
            if status in (ML_TASK_STATUS.WAITING, ML_TASK_STATUS.PROCESSING):
                continue
            if status == ML_TASK_STATUS.ERROR:
                if self.exception is not None:
                    raise self.exception
                else:
                    raise Exception("Unknown error during ML task execution")
            if status == ML_TASK_STATUS.TIMEOUT:
                raise Exception(f"Can't get answer in {self._timeout} seconds")
            if status == ML_TASK_STATUS.COMPLETE:
                return
            raise KeyError("Unknown task status")

    def result(self) -> DataFrame:
        """wait task is done and return result

        Returns:
            DataFrame: task result
        """
        self.wait()
        return self.dataframe

    def add_done_callback(self, fn: Callable) -> None:
        """need for compatability with concurrent.futures.Future interface"""
        pass



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Redshift cluster.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The port to use when connecting with the Redshift cluster.',
        'required': True,
        'label': 'Port'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Redshift cluster.',
        'required': True,
        'label': 'Database'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Redshift cluster.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Redshift cluster.',
        'required': True,
        'label': 'Password',
        'secret': True
    },
    schema={
        'type': ARG_TYPE.STR,
        'description': 'The schema in which objects are searched first.',
        'required': False,
        'label': 'Schema'
    },
    sslmode={
        'type': ARG_TYPE.STR,
        'description': 'The SSL mode that will be used for connection.',
        'required': False,
        'label': 'sslmode'
    }
)

connection_args_example = OrderedDict(
    host='examplecluster.abc123xyz789.us-west-1.redshift.amazonaws.com',
    port='5439',
    database='dev',
    user='awsuser',
    password='my_password'
)



from typing import List, Optional, Iterable
import pickle
import datetime as dt

from sqlalchemy.orm.attributes import flag_modified
import pandas as pd

from mindsdb_sql_parser import Select, Star, OrderBy

from mindsdb_sql_parser.ast import Identifier, BinaryOperation, Last, Constant, ASTNode
from mindsdb.integrations.utilities.query_traversal import query_traversal
from mindsdb.utilities.cache import get_cache

from mindsdb.interfaces.storage import db
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.config import config

from .last_query import LastQuery


class RunningQuery:
    """
    Query in progres
    """

    OBJECT_TYPE = "query"

    def __init__(self, record: db.Queries):
        self.record = record
        self.sql = record.sql
        self.database = record.database or config.get("default_project")

    def get_partitions(self, dn, step_call, query: Select) -> Iterable:
        """
        Gets chunks of data from data handler for executing them in next steps of the planner
        Check if datanode supports fetch with stream
        :param dn: datanode to execute query
        :param step_call: instance of StepCall to get some parameters from it
        :param query: AST query to execute
        :return: generator with query results
        """
        if hasattr(dn, "has_support_stream") and dn.has_support_stream():
            query2 = self.get_partition_query(step_call.current_step_num, query, stream=True)

            for df in dn.query_stream(query2, fetch_size=self.batch_size):
                max_track_value = self.get_max_track_value(df)
                yield df
                self.set_progress(max_track_value=max_track_value)

        else:
            while True:
                query2 = self.get_partition_query(step_call.current_step_num, query, stream=False)

                response = dn.query(query=query2, session=step_call.session)
                df = response.data_frame

                if df is None or len(df) == 0:
                    break

                max_track_value = self.get_max_track_value(df)
                yield df
                self.set_progress(max_track_value=max_track_value)

    def get_partition_query(self, step_num: int, query: Select, stream=False) -> Select:
        """
        Generate query for fetching the next partition
        It wraps query to
           select * from ({query})
           where {track_column} > {previous_value}
           order by track_column
           limit size {batch_size}
        And fill track_column, previous_value, batch_size

        If steam is true:
          - if track_column is defined:
             - don't add limit
          - else:
             - return user query without modifications
        """

        track_column = self.record.parameters.get("track_column")
        if track_column is None and stream:
            # if no track column for stream fetching: it is not resumable query, execute original query

            # check if it is first run of the query
            if self.record.processed_rows > 0:
                raise RuntimeError("Can't resume query without track_column")
            return query

        if not stream and track_column is None:
            raise ValueError("Track column is not defined")

        query = Select(
            targets=[Star()],
            from_table=query,
            order_by=[OrderBy(Identifier(track_column))],
        )
        if not stream:
            query.limit = Constant(self.batch_size)

        track_value = self.record.context.get("track_value")
        # is it different step?
        cur_step_num = self.record.context.get("step_num")
        if cur_step_num is not None and cur_step_num != step_num:
            # reset track_value
            track_value = None
            self.record.context["track_value"] = None
            self.record.context["step_num"] = step_num
            flag_modified(self.record, "context")
            db.session.commit()

        if track_value is not None:
            query.where = BinaryOperation(
                op=">",
                args=[Identifier(track_column), Constant(track_value)],
            )

        return query

    def get_info(self):
        record = self.record
        return {
            "id": record.id,
            "sql": record.sql,
            "database": record.database,
            "started_at": record.started_at,
            "finished_at": record.finished_at,
            "parameters": record.parameters,
            "context": record.context,
            "processed_rows": record.processed_rows,
            "error": record.error,
            "updated_at": record.updated_at,
        }

    def add_to_task(self):
        task_record = db.Tasks(
            company_id=ctx.company_id,
            user_class=ctx.user_class,
            object_type=self.OBJECT_TYPE,
            object_id=self.record.id,
        )
        db.session.add(task_record)
        db.session.commit()

    def remove_from_task(self):
        task = db.Tasks.query.filter(
            db.Tasks.object_type == self.OBJECT_TYPE,
            db.Tasks.object_id == self.record.id,
            db.Tasks.company_id == ctx.company_id,
        ).first()

        if task is not None:
            db.session.delete(task)
            db.session.commit()

    def set_params(self, params: dict):
        """
        Store parameters of the step which is about to be split into partitions
        """

        if "batch_size" not in params:
            params["batch_size"] = 1000

        self.record.parameters = params
        self.batch_size = self.record.parameters["batch_size"]
        db.session.commit()

    def get_max_track_value(self, df: pd.DataFrame) -> Optional[pd.DataFrame]:
        """
        return max value to use in `set_progress`.
        this function is called before execution substeps,
         `set_progress` function - after
        """
        if "track_column" in self.record.parameters:
            track_column = self.record.parameters["track_column"]
            return df[track_column].max()
        else:
            # stream mode
            return None

    def set_progress(self, processed_rows: int = None, max_track_value: int = None):
        """
        Store progres of the query, it is called after processing of batch
        """

        if processed_rows is not None and processed_rows > 0:
            self.record.processed_rows = self.record.processed_rows + processed_rows
            db.session.commit()

        if max_track_value is not None:
            cur_value = self.record.context.get("track_value")
            new_value = max_track_value
            if new_value is not None:
                if cur_value is None or new_value > cur_value:
                    self.record.context["track_value"] = new_value
                    flag_modified(self.record, "context")
            db.session.commit()

    def on_error(self, error: Exception, step_num: int, steps_data: dict):
        """
        Saves error of the query in database
        Also saves step data and current step num to be able to resume query
        """
        self.record.error = str(error)
        self.record.context["step_num"] = step_num
        flag_modified(self.record, "context")

        # save steps_data
        cache = get_cache("steps_data")
        data = pickle.dumps(steps_data, protocol=5)
        cache.set(str(self.record.id), data)

        db.session.commit()

    def mark_as_run(self):
        """
        Mark query as running and reset error of the query
        """
        if self.record.finished_at is not None:
            raise RuntimeError("The query already finished")

        if self.record.started_at is None:
            self.record.started_at = dt.datetime.now()
            db.session.commit()
        elif self.record.error is not None:
            self.record.error = None
            db.session.commit()
        else:
            raise RuntimeError("The query might be running already")

    def get_state(self) -> dict:
        """
        Returns stored state for resuming the query
        """
        cache = get_cache("steps_data")
        key = self.record.id
        data = cache.get(key)
        cache.delete(key)

        steps_data = pickle.loads(data)

        return {
            "step_num": self.record.context.get("step_num"),
            "steps_data": steps_data,
        }

    def finish(self):
        """
        Mark query as finished
        """

        self.record.finished_at = dt.datetime.now()
        db.session.commit()


class QueryContextController:
    IGNORE_CONTEXT = "<IGNORE>"

    def handle_db_context_vars(self, query: ASTNode, dn, session) -> tuple:
        """
        Check context variables in query and replace them with values.
        Should be used before exec query in database.

        Input:
        - query: input query
        - params are used to find current values of context variables
          - dn: datanode
          - session: mindsdb server session

        Returns:
         - query with replaced context variables
         - callback to call with result of the query. it is used to update context variables

        """
        context_name = self.get_current_context()

        l_query = LastQuery(query)
        if l_query.query is None:
            # no last keyword, exit
            return query, None

        if context_name == self.IGNORE_CONTEXT:
            # return with empty constants
            return l_query.query, None

        query_str = l_query.to_string()

        rec = self._get_context_record(context_name, query_str)

        if rec is None or len(rec.values) == 0:
            values = self._get_init_last_values(l_query, dn, session)
            if rec is None:
                self.__add_context_record(context_name, query_str, values)
                if context_name.startswith("job-if-"):
                    # add context for job also
                    self.__add_context_record(context_name.replace("job-if", "job"), query_str, values)
            else:
                rec.values = values
        else:
            values = rec.values

        db.session.commit()

        query_out = l_query.apply_values(values)

        def callback(df, columns_info):
            self._result_callback(l_query, context_name, query_str, df, columns_info)

        return query_out, callback

    def remove_lasts(self, query):
        def replace_lasts(node, **kwargs):
            # find last in where
            if isinstance(node, BinaryOperation):
                if isinstance(node.args[0], Identifier) and isinstance(node.args[1], Last):
                    node.args = [Constant(0), Constant(0)]
                    node.op = "="

        # find lasts
        query_traversal(query, replace_lasts)
        return query

    def _result_callback(
        self, l_query: LastQuery, context_name: str, query_str: str, df: pd.DataFrame, columns_info: list
    ):
        """
        This function handlers result from executed query and updates context variables with new values

        Input
        - l_query: LastQuery object
        - To identify context:
          - context_name: name of the context
          - query_str: rendered query to search in context table
        - result of the query
          - data: list of dicts
          - columns_info: list

        """
        if len(df) == 0:
            return

        values = {}
        # get max values
        for info in l_query.get_last_columns():
            target_idx = info["target_idx"]
            if target_idx is not None:
                # get by index
                col_name = columns_info[target_idx]["name"]
            else:
                col_name = info["column_name"]
                # get by name
            if col_name not in df:
                continue

            column_values = df[col_name].dropna()
            try:
                value = max(column_values)
            except (TypeError, ValueError):
                try:
                    # try to convert to float
                    value = max(map(float, column_values))
                except (TypeError, ValueError):
                    try:
                        # try to convert to str
                        value = max(map(str, column_values))
                    except (TypeError, ValueError):
                        continue

            if value is not None:
                values[info["table_name"]] = {info["column_name"]: value}

        self.__update_context_record(context_name, query_str, values)

    def drop_query_context(self, object_type: str, object_id: int = None):
        """
        Drop context for object
        :param object_type: type of the object
        :param object_id: id
        """

        context_name = self.gen_context_name(object_type, object_id)
        for rec in (
            db.session.query(db.QueryContext).filter_by(context_name=context_name, company_id=ctx.company_id).all()
        ):
            db.session.delete(rec)
        db.session.commit()

    def _get_init_last_values(self, l_query: LastQuery, dn, session) -> dict:
        """
        Gets current last values for query.
        Creates and executes query for it:
           'select <col> from <table> order by <col> desc limit 1"
        """
        last_values = {}
        for query, info in l_query.get_init_queries():
            response = dn.query(query=query, session=session)
            data = response.data_frame
            columns_info = response.columns

            if len(data) == 0:
                value = None
            else:
                row = list(data.iloc[0])

                idx = None
                for i, col in enumerate(columns_info):
                    if col["name"].upper() == info["column_name"].upper():
                        idx = i
                        break

                if idx is None or len(row) == 1:
                    value = row[0]
                else:
                    value = row[idx]

            if value is not None:
                last_values[info["table_name"]] = {info["column_name"]: value}

        return last_values

    # Context

    def get_current_context(self) -> str:
        """
        returns current context name
        """
        try:
            context_stack = ctx.context_stack or []
        except AttributeError:
            context_stack = []
        if len(context_stack) > 0:
            return context_stack[-1]
        else:
            return ""

    def set_context(self, object_type: str = None, object_id: int = None):
        """
        Updates current context name, using object name and id
        Previous context names are stored on lower levels of stack
        """
        try:
            context_stack = ctx.context_stack or []
        except AttributeError:
            context_stack = []
        context_stack.append(self.gen_context_name(object_type, object_id))
        ctx.context_stack = context_stack

    def release_context(self, object_type: str = None, object_id: int = None):
        """
        Removed current context (defined by object type and id) and restored previous one
        """
        try:
            context_stack = ctx.context_stack or []
        except AttributeError:
            context_stack = []
        if len(context_stack) == 0:
            return
        context_name = self.gen_context_name(object_type, object_id)
        if context_stack[-1] == context_name:
            context_stack.pop()
        ctx.context_stack = context_stack

    def gen_context_name(self, object_type: str, object_id: int) -> str:
        """
        Generated name of the context according to object type and name
        :return: context name
        """

        if object_type is None:
            return ""
        if object_id is not None:
            object_type += "-" + str(object_id)
        return object_type

    def get_context_vars(self, object_type: str, object_id: int) -> List[dict]:
        """
        Return variables stored in context (defined by object type and id)

        :return: list of all context variables related to context name how they stored in context table
        """
        context_name = self.gen_context_name(object_type, object_id)
        vars = []
        for rec in db.session.query(db.QueryContext).filter_by(context_name=context_name, company_id=ctx.company_id):
            if rec.values is not None:
                vars.append(rec.values)

        return vars

    # DB
    def _get_context_record(self, context_name: str, query_str: str) -> db.QueryContext:
        """
        Find and return record for context and query string
        """

        return (
            db.session.query(db.QueryContext)
            .filter_by(query=query_str, context_name=context_name, company_id=ctx.company_id)
            .first()
        )

    def __add_context_record(self, context_name: str, query_str: str, values: dict) -> db.QueryContext:
        """
        Creates record (for context and query string) with values and returns it
        """
        rec = db.QueryContext(query=query_str, context_name=context_name, company_id=ctx.company_id, values=values)
        db.session.add(rec)
        return rec

    def __update_context_record(self, context_name: str, query_str: str, values: dict):
        """
        Updates context record with new values
        """
        rec = self._get_context_record(context_name, query_str)
        rec.values = values
        db.session.commit()

    def get_query(self, query_id: int) -> RunningQuery:
        """
        Get running query by id
        """

        rec = db.Queries.query.filter(db.Queries.id == query_id, db.Queries.company_id == ctx.company_id).first()

        if rec is None:
            raise RuntimeError(f"Query not found: {query_id}")
        return RunningQuery(rec)

    def create_query(self, query: ASTNode, database: str = None) -> RunningQuery:
        """
        Create a new running query from AST query
        """

        # remove old queries
        remove_query = db.session.query(db.Queries).filter(
            db.Queries.company_id == ctx.company_id, db.Queries.finished_at < (dt.datetime.now() - dt.timedelta(days=1))
        )
        for rec in remove_query.all():
            self.get_query(rec.id).remove_from_task()
            db.session.delete(rec)

        rec = db.Queries(
            sql=str(query),
            database=database,
            company_id=ctx.company_id,
        )

        db.session.add(rec)
        db.session.commit()
        return RunningQuery(rec)

    def list_queries(self) -> List[dict]:
        """
        Get list of all running queries with metadata
        """

        query = db.session.query(db.Queries).filter(db.Queries.company_id == ctx.company_id)
        return [RunningQuery(record).get_info() for record in query]

    def cancel_query(self, query_id: int):
        """
        Cancels running query by id
        """
        rec = db.Queries.query.filter(db.Queries.id == query_id, db.Queries.company_id == ctx.company_id).first()
        if rec is None:
            raise RuntimeError(f"Query not found: {query_id}")

        self.get_query(rec.id).remove_from_task()

        # the query in progress will fail when it tries to update status
        db.session.delete(rec)
        db.session.commit()


query_context_controller = QueryContextController()



"""query_context

Revision ID: 6cb02dfd7f61
Revises: b0382f5be48d
Create Date: 2023-10-03 12:36:00.008731

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db # noqa


# revision identifiers, used by Alembic.
revision = '6cb02dfd7f61'
down_revision = 'b0382f5be48d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'query_context',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('company_id', sa.Integer(), nullable=True),
        sa.Column('query', sa.String(), nullable=False),
        sa.Column('context_name', sa.String(), nullable=False),
        sa.Column('values', sa.JSON(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():

    op.drop_table('query_context')
    # ### end Alembic commands ###
